import argparse
import glob
import hashlib
import importlib.metadata
import multiprocessing
import os
import shlex
import shutil
import sys
from typing import List, Optional

from buildbase import (
    Platform,
    add_path,
    build_sora,
    build_webrtc,
    cd,
    cmake_path,
    cmd,
    cmdcap,
    get_macos_osver,
    get_sora_info,
    get_webrtc_info,
    get_webrtc_platform,
    get_windows_osver,
    install_cmake,
    install_llvm,
    install_openh264,
    install_rootfs,
    install_sora_and_deps,
    install_webrtc,
    mkdir_p,
    read_version_file,
)
from pypath import get_python_include_dir, get_python_version

BASE_DIR = os.path.abspath(os.path.dirname(__file__))


def install_deps(
    platform: Platform,
    source_dir,
    build_dir,
    install_dir,
    debug,
    local_webrtc_build_dir: Optional[str],
    local_webrtc_build_args: List[str],
    local_sora_cpp_sdk_dir: Optional[str],
    local_sora_cpp_sdk_args: List[str],
):
    version = read_version_file("DEPS")

    # multistrap を使った sysroot の構築
    if platform.target.package_name in (
        "ubuntu-22.04_armv8_jetson",
        "raspberry-pi-os_armv8",
        "ubuntu-22.04_armv8",
        "ubuntu-24.04_armv8",
    ):
        conf = os.path.join("multistrap", f"{platform.target.package_name}.conf")
        # conf ファイルのハッシュ値をバージョンとする
        version_md5 = hashlib.md5(open(conf, "rb").read()).hexdigest()
        install_rootfs_args = {
            "version": version_md5,
            "version_file": os.path.join(install_dir, "rootfs.version"),
            "install_dir": install_dir,
            "conf": conf,
        }
        install_rootfs(**install_rootfs_args)

    # WebRTC
    webrtc_platform = get_webrtc_platform(platform)
    if local_webrtc_build_dir is None:
        install_webrtc_args = {
            "version": version["WEBRTC_BUILD_VERSION"],
            "version_file": os.path.join(install_dir, "webrtc.version"),
            "source_dir": source_dir,
            "install_dir": install_dir,
            "platform": webrtc_platform,
        }
        install_webrtc(**install_webrtc_args)
    else:
        build_webrtc_args = {
            "platform": webrtc_platform,
            "local_webrtc_build_dir": local_webrtc_build_dir,
            "local_webrtc_build_args": local_webrtc_build_args,
            "debug": debug,
        }
        build_webrtc(**build_webrtc_args)

    webrtc_info = get_webrtc_info(webrtc_platform, local_webrtc_build_dir, install_dir, debug)
    webrtc_version = read_version_file(webrtc_info.version_file)

    if (
        platform.build.os == "macos" or platform.build.os == "ubuntu"
    ) and local_webrtc_build_dir is None:
        # LLVM
        tools_url = webrtc_version["WEBRTC_SRC_TOOLS_URL"]
        tools_commit = webrtc_version["WEBRTC_SRC_TOOLS_COMMIT"]
        libcxx_url = webrtc_version["WEBRTC_SRC_THIRD_PARTY_LIBCXX_SRC_URL"]
        libcxx_commit = webrtc_version["WEBRTC_SRC_THIRD_PARTY_LIBCXX_SRC_COMMIT"]
        buildtools_url = webrtc_version["WEBRTC_SRC_BUILDTOOLS_URL"]
        buildtools_commit = webrtc_version["WEBRTC_SRC_BUILDTOOLS_COMMIT"]
        install_llvm_args = {
            "version": f"{tools_url}.{tools_commit}."
            f"{libcxx_url}.{libcxx_commit}."
            f"{buildtools_url}.{buildtools_commit}",
            "version_file": os.path.join(install_dir, "llvm.version"),
            "install_dir": install_dir,
            "tools_url": tools_url,
            "tools_commit": tools_commit,
            "libcxx_url": libcxx_url,
            "libcxx_commit": libcxx_commit,
            "buildtools_url": buildtools_url,
            "buildtools_commit": buildtools_commit,
        }
        install_llvm(**install_llvm_args)

    # CMake
    install_cmake_args = {
        "version": version["CMAKE_VERSION"],
        "version_file": os.path.join(install_dir, "cmake.version"),
        "source_dir": source_dir,
        "install_dir": install_dir,
        "platform": "",
        "ext": "tar.gz",
    }
    if platform.build.os == "windows" and platform.build.arch == "x86_64":
        install_cmake_args["platform"] = "windows-x86_64"
        install_cmake_args["ext"] = "zip"
    elif platform.build.os == "macos":
        install_cmake_args["platform"] = "macos-universal"
    elif platform.build.os == "ubuntu" and platform.build.arch == "x86_64":
        install_cmake_args["platform"] = "linux-x86_64"
    elif platform.build.os == "ubuntu" and platform.build.arch == "armv8":
        install_cmake_args["platform"] = "linux-aarch64"
    else:
        raise Exception("Failed to install CMake")
    install_cmake(**install_cmake_args)

    # Sora C++ SDK
    if local_sora_cpp_sdk_dir is None:
        install_sora_and_deps(
            version["SORA_CPP_SDK_VERSION"],
            version["BOOST_VERSION"],
            platform.target.package_name,
            source_dir,
            install_dir,
        )
    else:
        build_sora(
            platform.target.package_name,
            local_sora_cpp_sdk_dir,
            local_sora_cpp_sdk_args,
            debug,
            local_webrtc_build_dir,
        )

    if platform.build.os == "macos":
        add_path(os.path.join(install_dir, "cmake", "CMake.app", "Contents", "bin"))
    else:
        add_path(os.path.join(install_dir, "cmake", "bin"))

    if platform.build.os != "windows":
        # OpenH264
        install_openh264_args = {
            "version": version["OPENH264_VERSION"],
            "version_file": os.path.join(install_dir, "openh264.version"),
            "source_dir": source_dir,
            "install_dir": install_dir,
            "is_windows": platform.target.os == "windows",
        }
        install_openh264(**install_openh264_args)


AVAILABLE_TARGETS = [
    "windows_x86_64",
    "macos_arm64",
    "ubuntu-22.04_x86_64",
    "ubuntu-24.04_x86_64",
    "ubuntu-22.04_armv8",
    "ubuntu-24.04_armv8",
    "ubuntu-22.04_armv8_jetson",
    "raspberry-pi-os_armv8",
]


def _find_clang_binary(name: str) -> Optional[str]:
    if shutil.which(name) is not None:
        return name
    else:
        for n in range(50, 14, -1):
            if shutil.which(f"{name}-{n}") is not None:
                return f"{name}-{n}"
    return None


def _get_platform(target: str) -> Platform:
    if target == "windows_x86_64":
        platform = Platform("windows", get_windows_osver(), "x86_64")
    elif target == "macos_x86_64":
        platform = Platform("macos", get_macos_osver(), "x86_64")
    elif target == "macos_arm64":
        platform = Platform("macos", get_macos_osver(), "arm64")
    elif target == "ubuntu-22.04_x86_64":
        platform = Platform("ubuntu", "22.04", "x86_64")
    elif target == "ubuntu-24.04_x86_64":
        platform = Platform("ubuntu", "24.04", "x86_64")
    elif target == "ubuntu-22.04_armv8":
        platform = Platform("ubuntu", "22.04", "armv8")
    elif target == "ubuntu-24.04_armv8":
        platform = Platform("ubuntu", "24.04", "armv8")
    elif target == "ubuntu-22.04_armv8_jetson":
        platform = Platform("jetson", None, "armv8", "ubuntu-22.04")
    elif target == "raspberry-pi-os_armv8":
        platform = Platform("raspberry-pi-os", None, "armv8")
    else:
        raise Exception(f"Unknown target {target}")
    return platform


def _build(
    target: str,
    debug: bool,
    relwithdebinfo: bool,
    local_webrtc_build_dir: Optional[str],
    local_webrtc_build_args: List[str],
    local_sora_cpp_sdk_dir: Optional[str],
    local_sora_cpp_sdk_args: List[str],
):
    platform = _get_platform(target)

    source_dir = os.path.join(BASE_DIR, "_source", platform.target.package_name)
    build_dir = os.path.join(BASE_DIR, "_build", platform.target.package_name)
    install_dir = os.path.join(BASE_DIR, "_install", platform.target.package_name)
    mkdir_p(source_dir)
    mkdir_p(build_dir)
    mkdir_p(install_dir)

    with cd(BASE_DIR):
        install_deps(
            platform,
            source_dir,
            build_dir,
            install_dir,
            debug,
            local_webrtc_build_dir,
            local_webrtc_build_args,
            local_sora_cpp_sdk_dir,
            local_sora_cpp_sdk_args,
        )

        configuration = "Release"
        if debug:
            configuration = "Debug"
        if relwithdebinfo:
            configuration = "RelWithDebInfo"

        webrtc_platform = get_webrtc_platform(platform)
        webrtc_info = get_webrtc_info(webrtc_platform, local_webrtc_build_dir, install_dir, debug)

        sora_info = get_sora_info(webrtc_platform, local_sora_cpp_sdk_dir, install_dir, debug)

        cmake_args = []
        cmake_args.append(f"-DCMAKE_BUILD_TYPE={configuration}")
        cmake_args.append(f"-DTARGET_OS={platform.target.os}")
        # Raspberry Pi OS の場合は sora-sdk-rpi パッケージからバージョンを取得する
        if platform.target.os == "raspberry-pi-os":
            cmake_args.append(
                f"-DSORA_PYTHON_SDK_VERSION={importlib.metadata.version('sora-sdk-rpi')}"
            )
        else:
            cmake_args.append(f"-DSORA_PYTHON_SDK_VERSION={importlib.metadata.version('sora-sdk')}")
        cmake_args.append(f"-DBOOST_ROOT={cmake_path(sora_info.boost_install_dir)}")
        cmake_args.append(f"-DWEBRTC_INCLUDE_DIR={cmake_path(webrtc_info.webrtc_include_dir)}")
        cmake_args.append(f"-DWEBRTC_LIBRARY_DIR={cmake_path(webrtc_info.webrtc_library_dir)}")
        cmake_args.append(f"-DSORA_DIR={cmake_path(sora_info.sora_install_dir)}")
        cmake_args.append(f"-DOPENH264_DIR={cmake_path(os.path.join(install_dir, 'openh264'))}")
        python_version = get_python_version()
        cmake_args.append(f"-DPython_INCLUDE_DIR={get_python_include_dir(python_version)}")
        cmake_args.append(f"-DPython_EXECUTABLE={cmake_path(sys.executable)}")

        if platform.target.os == "ubuntu":
            # クロスコンパイルの設定。
            # 本来は toolchain ファイルに書く内容

            if platform.build.arch == "armv8":
                # ビルド環境が armv8 の場合は libwebrtc のバイナリが使えないのでローカルの clang を利用する
                cmake_args += [
                    "-DCMAKE_C_COMPILER=clang-19",
                    "-DCMAKE_CXX_COMPILER=clang++-19",
                ]
            else:
                cmake_args += [
                    f"-DCMAKE_C_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang')}",
                    f"-DCMAKE_CXX_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang++')}",
                ]
            cmake_args += [
                f"-DLIBCXX_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxx_dir, 'include'))}",
                f"-DLIBCXXABI_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxxabi_dir, 'include'))}",
            ]

            if platform.build.arch != platform.target.arch:
                sysroot = os.path.join(install_dir, "rootfs")
                python_version = get_python_version()
                python_version_short = ".".join(python_version.split(".")[:2])
                cmake_args += [
                    "-DCMAKE_SYSTEM_NAME=Linux",
                    "-DCMAKE_SYSTEM_PROCESSOR=aarch64",
                    "-DCMAKE_C_COMPILER_TARGET=aarch64-linux-gnu",
                    "-DCMAKE_CXX_COMPILER_TARGET=aarch64-linux-gnu",
                    f"-DCMAKE_FIND_ROOT_PATH={sysroot}",
                    "-DCMAKE_FIND_ROOT_PATH_MODE_PROGRAM=NEVER",
                    "-DCMAKE_FIND_ROOT_PATH_MODE_LIBRARY=BOTH",
                    "-DCMAKE_FIND_ROOT_PATH_MODE_INCLUDE=BOTH",
                    "-DCMAKE_FIND_ROOT_PATH_MODE_PACKAGE=BOTH",
                    f"-DCMAKE_SYSROOT={sysroot}",
                    f"-DNB_SUFFIX=.cpython-{python_version_short.replace('.', '')}-aarch64-linux-gnu.so",
                ]
        elif platform.target.os == "macos":
            sysroot = cmdcap(["xcrun", "--sdk", "macosx", "--show-sdk-path"])
            cmake_args += [
                "-DCMAKE_SYSTEM_PROCESSOR=arm64",
                "-DCMAKE_OSX_ARCHITECTURES=arm64",
                f"-DCMAKE_C_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang')}",
                "-DCMAKE_C_COMPILER_TARGET=aarch64-apple-darwin",
                f"-DCMAKE_CXX_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang++')}",
                "-DCMAKE_CXX_COMPILER_TARGET=aarch64-apple-darwin",
                f"-DCMAKE_SYSROOT={sysroot}",
                f"-DLIBCXX_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxx_dir, 'include'))}",
                f"-DLIBCXXABI_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxxabi_dir, 'include'))}",
            ]
        elif platform.target.os == "jetson":
            sysroot = os.path.join(install_dir, "rootfs")
            cmake_args += [
                "-DCMAKE_SYSTEM_NAME=Linux",
                "-DCMAKE_SYSTEM_PROCESSOR=aarch64",
                f"-DCMAKE_C_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang')}",
                "-DCMAKE_C_COMPILER_TARGET=aarch64-linux-gnu",
                f"-DCMAKE_CXX_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang++')}",
                "-DCMAKE_CXX_COMPILER_TARGET=aarch64-linux-gnu",
                f"-DCMAKE_FIND_ROOT_PATH={sysroot}",
                "-DCMAKE_FIND_ROOT_PATH_MODE_PROGRAM=NEVER",
                "-DCMAKE_FIND_ROOT_PATH_MODE_LIBRARY=BOTH",
                "-DCMAKE_FIND_ROOT_PATH_MODE_INCLUDE=BOTH",
                "-DCMAKE_FIND_ROOT_PATH_MODE_PACKAGE=BOTH",
                f"-DCMAKE_SYSROOT={sysroot}",
                f"-DLIBCXX_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxx_dir, 'include'))}",
                f"-DLIBCXXABI_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxxabi_dir, 'include'))}",
                f"-DPython_ROOT_DIR={cmake_path(os.path.join(sysroot, 'usr', 'include', 'python3.10'))}",
                "-DNB_SUFFIX=.cpython-310-aarch64-linux-gnu.so",
            ]
        elif platform.target.os == "raspberry-pi-os":
            sysroot = os.path.join(install_dir, "rootfs")
            python_version = get_python_version()
            python_version_short = ".".join(python_version.split(".")[:2])
            cmake_args += [
                "-DCMAKE_SYSTEM_NAME=Linux",
                "-DCMAKE_SYSTEM_PROCESSOR=aarch64",
                f"-DCMAKE_C_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang')}",
                "-DCMAKE_C_COMPILER_TARGET=aarch64-linux-gnu",
                f"-DCMAKE_CXX_COMPILER={os.path.join(webrtc_info.clang_dir, 'bin', 'clang++')}",
                "-DCMAKE_CXX_COMPILER_TARGET=aarch64-linux-gnu",
                f"-DCMAKE_FIND_ROOT_PATH={sysroot}",
                "-DCMAKE_FIND_ROOT_PATH_MODE_PROGRAM=NEVER",
                "-DCMAKE_FIND_ROOT_PATH_MODE_LIBRARY=BOTH",
                "-DCMAKE_FIND_ROOT_PATH_MODE_INCLUDE=BOTH",
                "-DCMAKE_FIND_ROOT_PATH_MODE_PACKAGE=BOTH",
                f"-DCMAKE_SYSROOT={sysroot}",
                f"-DLIBCXX_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxx_dir, 'include'))}",
                f"-DLIBCXXABI_INCLUDE_DIR={cmake_path(os.path.join(webrtc_info.libcxxabi_dir, 'include'))}",
                f"-DNB_SUFFIX=.cpython-{python_version_short.replace('.', '')}-aarch64-linux-gnu.so",
            ]

        # Windows 以外の、クロスコンパイルでない環境では pyi ファイルを生成する
        if (
            platform.target.os != "windows"
            and platform.build.package_name == platform.target.package_name
        ):
            cmake_args.append("-DSORA_GEN_PYI=ON")

        sora_src_dir = os.path.join("src", "sora_sdk")
        sora_build_dir = os.path.join(build_dir, "sora_sdk")
        if platform.target.os == "windows":
            sora_build_target_dir = os.path.join(build_dir, "sora_sdk", configuration)
        else:
            sora_build_target_dir = os.path.join(build_dir, "sora_sdk")

        mkdir_p(sora_build_dir)
        with cd(sora_build_dir):
            cmd(["cmake", BASE_DIR, *cmake_args])
            cmd(
                [
                    "cmake",
                    "--build",
                    ".",
                    "--config",
                    configuration,
                    f"-j{multiprocessing.cpu_count()}",
                ]
            )

        for file in os.listdir(sora_src_dir):
            if file.startswith("sora_sdk_ext.") and (
                file.endswith(".so") or file.endswith(".dylib") or file.endswith(".pyd")
            ):
                os.remove(os.path.join(sora_src_dir, file))

        for file in os.listdir(sora_build_target_dir):
            if file.startswith("sora_sdk_ext.") and (
                file.endswith(".so") or file.endswith(".dylib") or file.endswith(".pyd")
            ):
                shutil.copyfile(
                    os.path.join(sora_build_target_dir, file), os.path.join(sora_src_dir, file)
                )
            if file in ("sora_sdk_ext.pyi", "py.typed"):
                shutil.copyfile(
                    os.path.join(sora_build_target_dir, file), os.path.join(sora_src_dir, file)
                )

        if platform.target.os == "raspberry-pi-os":
            # libcamerac.so を sora_sdk_ext.*.so と同じディレクトリにコピーする
            libcamerac_so = os.path.join(sora_info.sora_install_dir, "lib", "libcamerac.so")
            shutil.copyfile(libcamerac_so, os.path.join(sora_src_dir, "libcamerac.so"))


def _format(
    clang_format_path: Optional[str] = None,
    skip_clang_format: bool = False,
    skip_ruff: bool = False,
):
    # C++ ファイルのフォーマット
    if not skip_clang_format:
        if clang_format_path is None:
            clang_format_path = _find_clang_binary("clang-format")
        if clang_format_path is None:
            print("Warning: clang-format not found. Skipping C++ formatting.")
        else:
            patterns = [
                "src/**/*.h",
                "src/**/*.cpp",
            ]
            target_files = []
            for pattern in patterns:
                files = glob.glob(pattern, recursive=True)
                target_files.extend(files)
            if target_files:
                print(f"Formatting {len(target_files)} C++ files...")
                cmd([clang_format_path, "-i"] + target_files)

    # Python ファイルのフォーマット
    if not skip_ruff:
        print("Running Python formatting with ruff...")
        try:
            cmd(["uv", "run", "ruff", "format"])
        except Exception as e:
            print(f"Formatting failed: {e}")


def main():
    parser = argparse.ArgumentParser()
    sp = parser.add_subparsers(dest="command", required=True)

    # build サブコマンド
    bp = sp.add_parser("build")
    bp.add_argument("target", choices=AVAILABLE_TARGETS)
    bp.add_argument("--debug", action="store_true")
    bp.add_argument("--relwithdebinfo", action="store_true")
    bp.add_argument("--local-webrtc-build-dir", type=os.path.abspath)
    bp.add_argument("--local-webrtc-build-args", default="", type=shlex.split)
    bp.add_argument("--local-sora-cpp-sdk-dir", type=os.path.abspath)
    bp.add_argument("--local-sora-cpp-sdk-args", default="", type=shlex.split)

    # format サブコマンド
    fp = sp.add_parser("format")
    fp.add_argument("--clang-format-path", type=str, default=None)
    fp.add_argument(
        "--skip-clang-format", action="store_true", help="Skip C++ formatting with clang-format"
    )
    fp.add_argument("--skip-ruff", action="store_true", help="Skip Python formatting with ruff")

    args = parser.parse_args()

    if args.command == "build":
        _build(
            target=args.target,
            debug=args.debug,
            relwithdebinfo=args.relwithdebinfo,
            local_webrtc_build_dir=args.local_webrtc_build_dir,
            local_webrtc_build_args=args.local_webrtc_build_args,
            local_sora_cpp_sdk_dir=args.local_sora_cpp_sdk_dir,
            local_sora_cpp_sdk_args=args.local_sora_cpp_sdk_args,
        )
    elif args.command == "format":
        _format(
            clang_format_path=args.clang_format_path,
            skip_clang_format=args.skip_clang_format,
            skip_ruff=args.skip_ruff,
        )


if __name__ == "__main__":
    main()



import json
import threading
import time
from threading import Event
from typing import Any, Optional

import numpy
from conftest import Settings

from sora_sdk import (
    Sora,
    SoraAudioFrameTransformer,
    SoraAudioSource,
    SoraMediaTrack,
    SoraTransformableAudioFrame,
    SoraTransformableVideoFrame,
    SoraVideoFrameTransformer,
    SoraVideoSource,
)


class SendonlyEncodedTransform:
    def __init__(
        self,
        settings: Settings,
        metadata: dict[str, Any] | None = None,
        jwt_private_claims: dict[str, Any] | None = None,
    ):
        self._signaling_urls: list[str] = settings.signaling_urls
        self._channel_id: str = settings.channel_id

        self._connection_id: str

        if jwt_private_claims is not None:
            access_token = settings.access_token(**jwt_private_claims)
        else:
            access_token = settings.access_token()

        # secret が設定されていない場合は access_token が存在しない
        if access_token is not None:
            if metadata is not None:
                # metadata が設定されている場合は access_token を追加する
                metadata.update({"access_token": access_token})
            else:
                # metadata が設定されていない場合は access_token のみを metadata に設定する
                metadata = {"access_token": access_token}

        # 接続した
        self._connected: Event = Event()
        # 終了
        self._disconnected = Event()

        self._audio_channels: int = 1
        self._audio_sample_rate: int = 16000

        self._video_width: int = 960
        self._video_height: int = 540

        self._sora = Sora()

        self._fake_audio_thread: Optional[threading.Thread] = None
        self._fake_video_thread: Optional[threading.Thread] = None

        self._audio_source: Optional[SoraAudioSource] = None
        self._audio_source = self._sora.create_audio_source(
            self._audio_channels, self._audio_sample_rate
        )

        self._video_source: Optional[SoraVideoSource] = None
        self._video_source = self._sora.create_video_source()

        # Audio 向けの Encoded Transformer
        self._audio_transformer = SoraAudioFrameTransformer()
        # Audio のエンコードフレームを受け取るコールバック関数を on_transform に設定
        self._audio_transformer.on_transform = self._on_audio_transform

        # Video 向けの Encoded Transformer
        self._video_transformer = SoraVideoFrameTransformer()
        # Video のエンコードフレームを受け取るコールバック関数を on_transform に設定
        self._video_transformer.on_transform = self._on_video_transform

        self._connection = self._sora.create_connection(
            signaling_urls=self._signaling_urls,
            role="sendonly",
            channel_id=self._channel_id,
            metadata=metadata,
            audio=True,
            video=True,
            audio_source=self._audio_source,
            video_source=self._video_source,
            audio_frame_transformer=self._audio_transformer,
            video_frame_transformer=self._video_transformer,
        )

        self._connection.on_set_offer = self._on_set_offer
        self._connection.on_notify = self._on_notify
        self._connection.on_disconnect = self._on_disconnect

        self._is_called_on_audio_transform = False
        self._is_called_on_video_transform = False

    def connect(self):
        self._connection.connect()

        self._fake_audio_thread = threading.Thread(target=self._fake_audio_loop, daemon=True)
        self._fake_audio_thread.start()

        self._fake_video_thread = threading.Thread(target=self._fake_video_loop, daemon=True)
        self._fake_video_thread.start()

        # _connected が set されるまで 30 秒待つ
        assert self._connected.wait(30)

        return self

    def disconnect(self):
        self._connection.disconnect()

    def get_stats(self):
        raw_stats = self._connection.get_stats()
        stats = json.loads(raw_stats)
        return stats

    @property
    def is_called_on_audio_transform(self):
        return self._is_called_on_audio_transform

    @property
    def is_called_on_video_transform(self):
        return self._is_called_on_video_transform

    def _fake_audio_loop(self):
        while not self._disconnected.is_set():
            time.sleep(0.02)
            if self._audio_source is not None:
                self._audio_source.on_data(numpy.zeros((320, 1), dtype=numpy.int16))

    def _fake_video_loop(self):
        while not self._disconnected.is_set():
            time.sleep(1.0 / 30)
            if self._video_source is not None:
                self._video_source.on_captured(
                    numpy.zeros((self._video_height, self._video_width, 3), dtype=numpy.uint8)
                )

    def _on_set_offer(self, raw_offer):
        offer = json.loads(raw_offer)
        if offer["type"] == "offer":
            self._connection_id = offer["connection_id"]
            print(f"Received 'Offer': connection_id={self._connection_id}")

    def _on_notify(self, raw_message):
        message = json.loads(raw_message)
        if (
            message["type"] == "notify"
            and message["event_type"] == "connection.created"
            and message["connection_id"] == self._connection_id
        ):
            print(f"Connected Sora: connection_id={self._connection_id}")
            self._connected.set()

    def _on_disconnect(self, error_code, message):
        print(f"Disconnected Sora: error_code='{error_code}' message='{message}'")
        self._disconnected.set()
        self._connected.clear()

        if self._fake_audio_thread is not None:
            self._fake_audio_thread.join(timeout=10)

        if self._fake_video_thread is not None:
            self._fake_video_thread.join(timeout=10)

    def _on_audio_transform(self, frame: SoraTransformableAudioFrame):
        # この実装が Encoded Transform を利用する上での基本形となる

        # frame からエンコードされたフレームデータを取得する
        # 戻り値は numpy.ndarray になっている
        new_data = frame.get_data()

        # "sora" という文字列を new_data の後ろに追加
        new_data = numpy.append(new_data, numpy.frombuffer(b"sora", dtype=numpy.uint8))

        self._is_called_on_audio_transform = True

        # ここで new_data の末尾にデータをつける new_data を暗号化するなど任意の処理を実装する

        # 加工したフレームデータで frame の フレームデータを入れ替える
        frame.set_data(new_data)
        self._audio_transformer.enqueue(frame)

    def _on_video_transform(self, frame: SoraTransformableVideoFrame):
        # この実装が Encoded Transform を利用する上での基本形となる

        # frame からエンコードされたフレームデータを取得する
        # 戻り値は numpy.ndarray になっている
        new_data = frame.get_data()

        # "sora" という文字列を new_data の後ろに追加
        new_data = numpy.append(new_data, numpy.frombuffer(b"sora", dtype=numpy.uint8))

        self._is_called_on_video_transform = True

        # ここで new_data の末尾にデータをつける new_data を暗号化するなど任意の処理を実装する

        # 加工したフレームデータで frame の フレームデータを入れ替える
        frame.set_data(new_data)
        self._video_transformer.enqueue(frame)


class RecvonlyEncodedTransform:
    def __init__(
        self,
        settings: Settings,
        metadata: dict[str, Any] | None = None,
        jwt_private_claims: dict[str, Any] | None = None,
    ):
        self._signaling_urls: list[str] = settings.signaling_urls
        self._channel_id: str = settings.channel_id

        if jwt_private_claims is not None:
            access_token = settings.access_token(**jwt_private_claims)
        else:
            access_token = settings.access_token()

        # secret が設定されていない場合は access_token が存在しない
        if access_token is not None:
            if metadata is not None:
                # metadata が設定されている場合は access_token を追加する
                metadata.update({"access_token": access_token})
            else:
                # metadata が設定されていない場合は access_token のみを metadata に設定する
                metadata = {"access_token": access_token}

        self._connection_id: str

        # 接続した
        self._connected: Event = Event()
        # 終了
        self._disconnected = Event()

        self._audio_output_frequency: int = 24000
        self._audio_output_channels: int = 1

        self._sora = Sora()

        self._connection = self._sora.create_connection(
            signaling_urls=self._signaling_urls,
            role="recvonly",
            channel_id=self._channel_id,
            metadata=metadata,
            audio=True,
            video=True,
        )

        self._connection.on_set_offer = self._on_set_offer
        self._connection.on_notify = self._on_notify
        self._connection.on_disconnect = self._on_disconnect

        self._connection.on_track = self._on_track

        self._is_called_on_audio_transform = False
        self._is_called_on_video_transform = False

    def connect(self):
        self._connection.connect()

        # _connected が set されるまで 30 秒待つ
        assert self._connected.wait(30)

        return self

    def disconnect(self):
        self._connection.disconnect()

    def get_stats(self):
        raw_stats = self._connection.get_stats()
        stats = json.loads(raw_stats)
        return stats

    @property
    def is_called_on_audio_transform(self):
        return self._is_called_on_audio_transform

    @property
    def is_called_on_video_transform(self):
        return self._is_called_on_video_transform

    def _on_set_offer(self, raw_offer):
        offer = json.loads(raw_offer)
        if offer["type"] == "offer":
            self._connection_id = offer["connection_id"]
            print(f"Received 'Offer': connection_id={self._connection_id}")

    def _on_notify(self, raw_message):
        message = json.loads(raw_message)
        if (
            message["type"] == "notify"
            and message["event_type"] == "connection.created"
            and message["connection_id"] == self._connection_id
        ):
            print(f"Connected Sora: connection_id={self._connection_id}")
            self._connected.set()

    def _on_disconnect(self, error_code, message):
        print(f"Disconnected Sora: error_code='{error_code}' message='{message}'")
        self._disconnected.set()
        self._connected.clear()

    def _on_track(self, track: SoraMediaTrack) -> None:
        if track.kind == "audio":
            # Audio 向けの Encoded Transformer
            self._audio_transformer = SoraAudioFrameTransformer()
            # Audio のエンコードフレームを受け取るコールバック関数を on_transform に設定
            self._audio_transformer.on_transform = self._on_audio_transform
            # Encoded Transformer を RTPReceiver に設定する
            track.set_frame_transformer(self._audio_transformer)
        if track.kind == "video":
            # Video 向けの Encoded Transformer
            self._video_transformer = SoraVideoFrameTransformer()
            # Video のエンコードフレームを受け取るコールバック関数を on_transform に設定
            self._video_transformer.on_transform = self._on_video_transform
            # Encoded Transformer を SoraMediaTrack に設定する
            track.set_frame_transformer(self._video_transformer)

    def _on_audio_transform(self, frame: SoraTransformableAudioFrame):
        # この実装が Encoded Transform を利用する上での基本形となる

        # frame からエンコードされたフレームデータを取得する
        # 戻り値は ArrayLike になっている
        new_data = frame.get_data()

        # ここで new_data の末尾にデータをつける new_data を暗号化するなど任意の処理を実装する

        # ArrayLike を numpy.uint8 のバイト列に変換する
        new_data = numpy.asarray(new_data, dtype=numpy.uint8)

        # 後ろ4バイトを取得する
        removed_data = new_data[-4:]

        assert b"sora" == removed_data.tobytes()

        # 後ろ4バイトを取り除く
        new_data = new_data[:-4]

        self._is_called_on_audio_transform = True

        # 加工したフレームデータで frame の フレームデータを入れ替える
        frame.set_data(new_data)
        self._audio_transformer.enqueue(frame)

    def _on_video_transform(self, frame: SoraTransformableVideoFrame):
        # この実装が Encoded Transform を利用する上での基本形となる
        # frame からエンコードされたフレームデータを取得する
        # 戻り値は ArrayLike になっている
        new_data = frame.get_data()

        # ここで new_data の末尾にデータをつける new_data を暗号化するなど任意の処理を実装する

        # ArrayLike を numpy.uint8 のバイト列に変換する
        new_data = numpy.asarray(new_data, dtype=numpy.uint8)

        # 後ろ4バイトを取得する
        removed_data = new_data[-4:]

        assert b"sora" == removed_data.tobytes()

        # 後ろ4バイトを取り除く
        new_data = new_data[:-4]

        self._is_called_on_video_transform = True

        # 加工したフレームデータで frame の フレームデータを入れ替える
        frame.set_data(new_data)
        self._video_transformer.enqueue(frame)


def test_encoded_transform(settings):
    sendonly = SendonlyEncodedTransform(settings)
    sendonly.connect()

    recvonly = RecvonlyEncodedTransform(settings)
    recvonly.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    recvonly_stats = recvonly.get_stats()

    sendonly.disconnect()
    recvonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(
        s for s in sendonly_stats if s.get("type") == "codec" and s.get("mimeType") == "audio/opus"
    )
    assert sendonly_codec_stats["mimeType"] == "audio/opus"

    sendonly_codec_stats = next(
        s for s in sendonly_stats if s.get("type") == "codec" and s.get("mimeType") == "video/VP9"
    )
    assert sendonly_codec_stats["mimeType"] == "video/VP9"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "audio"
    )
    # audio には encoderImplementation が無い
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "video"
    )
    # video には encoderImplementation が無い
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # codec が無かったら StopIteration 例外が上がる
    recvonly_codec_stats = next(
        s for s in recvonly_stats if s.get("type") == "codec" and s.get("mimeType") == "audio/opus"
    )
    assert recvonly_codec_stats["mimeType"] == "audio/opus"

    recvonly_codec_stats = next(
        s for s in recvonly_stats if s.get("type") == "codec" and s.get("mimeType") == "video/VP9"
    )
    assert recvonly_codec_stats["mimeType"] == "video/VP9"

    # inbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(
        s for s in recvonly_stats if s.get("type") == "inbound-rtp" and s.get("kind") == "audio"
    )
    # audio には encoderImplementation が無い
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0

    # inbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(
        s for s in recvonly_stats if s.get("type") == "inbound-rtp" and s.get("kind") == "video"
    )
    # video には encoderImplementation が無い
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0

    # on_transform が呼ばれていることを確認
    assert sendonly.is_called_on_audio_transform is True
    assert sendonly.is_called_on_video_transform is True

    assert recvonly.is_called_on_audio_transform is True
    assert recvonly.is_called_on_video_transform is True



import os
import time

import pytest
from client import SoraClient, SoraRole
from simulcast import default_video_bit_rate, expect_target_bitrate

pytestmark = pytest.mark.skipif(
    os.environ.get("APPLE_VIDEO_TOOLBOX") is None, reason="Apple Video Toolbox でのみ実行する"
)


@pytest.mark.parametrize(
    "video_codec_type",
    ["H264", "H265"],
)
def test_apple_video_toolbox_sendonly(settings, video_codec_type):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
    )
    sendonly.connect(fake_video=True)

    time.sleep(5)

    assert sendonly.connect_message is not None
    assert sendonly.connect_message["channel_id"] == settings.channel_id
    assert "video" in sendonly.connect_message
    assert sendonly.connect_message["video"]["codec_type"] == video_codec_type

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    # H.264 が採用されているかどうか確認する
    assert codec_stats["mimeType"] == f"video/{video_codec_type}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == "VideoToolbox"
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0


@pytest.mark.parametrize(
    "video_codec_type",
    ["H264", "H265"],
)
def test_apple_video_toolbox_sendonly_recvonly(settings, video_codec_type):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
    )
    sendonly.connect(fake_video=True)

    recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
    )
    recvonly.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    recvonly_stats = recvonly.get_stats()

    sendonly.disconnect()
    recvonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    # 指定した video_codec_type が採用されているかどうか確認する
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == "VideoToolbox"
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # codec が無かったら StopIteration 例外が上がる
    recvonly_codec_stats = next(s for s in recvonly_stats if s.get("type") == "codec")
    # 指定した video_codec_type が採用されているかどうか確認する
    assert recvonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(s for s in recvonly_stats if s.get("type") == "inbound-rtp")
    assert inbound_rtp_stats["decoderImplementation"] == "VideoToolbox"
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "expected_implementation",
        "video_width",
        "video_height",
        "simulcast_count",
    ),
    [
        # 1080p
        ("H264", "VideoToolbox", 1920, 1080, 3),
        ("H265", "VideoToolbox", 1920, 1080, 3),
        # 720p
        ("H264", "VideoToolbox", 1280, 720, 3),
        ("H265", "VideoToolbox", 1280, 720, 3),
        # 540p
        ("H264", "VideoToolbox", 960, 540, 3),
        ("H265", "VideoToolbox", 960, 540, 3),
        # 360p
        # ("H264", "VideoToolbox", 640, 360, 2),
        # ("H265", "VideoToolbox", 640, 360, 2),
        # 270p
        # ("H264", "VideoToolbox", 480, 270, 2),
        # ("H265", "VideoToolbox", 480, 270, 2),
        # 180p
        # ("H264", "VideoToolbox", 320, 180, 1),
        # ("H265", "VideoToolbox", 320, 180, 1),
        # 135p
        # ("H265", "VideoToolbox", 240, 135, 1),
    ],
)
def test_apple_video_toolbox_simulcast(
    settings,
    video_codec_type,
    expected_implementation,
    video_width,
    video_height,
    simulcast_count,
):
    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        simulcast=True,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_bit_rate=video_bit_rate,
        video_width=video_width,
        video_height=video_height,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # "type": "answer" の SDP で Simulcast があるかどうか
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert "a=simulcast:send r0;r1;r2" in sendonly.answer_message["sdp"]

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "video"
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 3

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert "qualityLimitationReason" in s
        assert "qualityLimitationDurations" in s

        # qualityLimitationReason が none で無い場合は安定したテストができない
        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["rid"] == f"r{i}"
        # simulcast_count が 2 の場合、rid r2 の bytesSent/packetsSent は 0 or 1 になる
        # simulcast_count が 1 の場合、rid r2 と r1 の bytesSent/packetsSent は 0 or 1 になる
        if i < simulcast_count:
            # 1 本になると simulcastEncodingAdapter がなくなる
            if simulcast_count > 1:
                assert "SimulcastEncoderAdapter" in s["encoderImplementation"]
            assert expected_implementation in s["encoderImplementation"]

            assert s["bytesSent"] > 500
            assert s["packetsSent"] > 5

            assert s["targetBitrate"] >= expect_target_bitrate(
                video_codec_type, s["frameWidth"], s["frameHeight"]
            )

            print(
                s["rid"],
                video_codec_type,
                expected_implementation,
                video_bit_rate * 1000,
                s["targetBitrate"],
                s["frameWidth"],
                s["frameHeight"],
                s["bytesSent"],
                s["packetsSent"],
            )

        else:
            # 本来は 0 なのだが、simulcast_count が 1 の場合、
            # packetSent が 0 ではなく 1 や 2 になる場合がある
            # byteSent は 0
            assert s["bytesSent"] == 0
            assert s["packetsSent"] <= 2
            print(
                s["rid"],
                video_codec_type,
                s["bytesSent"],
                s["packetsSent"],
            )


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "encoder_implementation",
        "video_bit_rate",
        "video_width",
        "video_height",
    ),
    [
        # どうやら scaleResolutionDownTo を指定すると規定されたテーブルのビットレートでは足りない
        # なので少しかさまししている
        ("H264", "VideoToolbox", 200 * 3, 320, 180),
        ("H265", "VideoToolbox", 142 * 3, 320, 180),
    ],
)
def test_apple_video_toolbox_simulcast_authz_scale_resolution_to(
    settings,
    video_codec_type,
    encoder_implementation,
    video_bit_rate,
    video_width,
    video_height,
):
    simulcast_encodings = [
        {
            "rid": "r0",
            "active": True,
            "scaleResolutionDownTo": {"maxWidth": video_width, "maxHeight": video_height},
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r1",
            "active": True,
            "scaleResolutionDownTo": {"maxWidth": video_width, "maxHeight": video_height},
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r2",
            "active": True,
            "scaleResolutionDownTo": {"maxWidth": video_width, "maxHeight": video_height},
            "scalabilityMode": "L1T1",
        },
    ]

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_bit_rate=video_bit_rate,
        jwt_private_claims={
            "video": True,
            "video_codec_type": video_codec_type,
            "video_bit_rate": video_bit_rate,
            "simulcast": True,
            "simulcast_encodings": simulcast_encodings,
        },
        video_width=video_width,
        video_height=video_height,
    )
    sendonly.connect(fake_video=True)

    time.sleep(5)

    # "type": "offer" の SDP で Simulcast があるかどうか
    assert sendonly.offer_message is not None
    assert sendonly.offer_message["sdp"] is not None
    assert video_codec_type in sendonly.offer_message["sdp"]
    assert "a=simulcast:recv r0;r1;r2" in sendonly.offer_message["sdp"]

    assert "encodings" in sendonly.offer_message
    assert len(sendonly.offer_message["encodings"]) == 3

    assert sendonly.offer_message["encodings"][0]["rid"] == simulcast_encodings[0]["rid"]
    assert sendonly.offer_message["encodings"][1]["rid"] == simulcast_encodings[1]["rid"]
    assert sendonly.offer_message["encodings"][2]["rid"] == simulcast_encodings[2]["rid"]

    assert sendonly.offer_message["encodings"][0]["active"] == simulcast_encodings[0]["active"]
    assert sendonly.offer_message["encodings"][1]["active"] == simulcast_encodings[1]["active"]
    assert sendonly.offer_message["encodings"][2]["active"] == simulcast_encodings[2]["active"]

    assert (
        sendonly.offer_message["encodings"][0]["scaleResolutionDownTo"]["maxWidth"]
        == simulcast_encodings[0]["scaleResolutionDownTo"]["maxWidth"]
    )
    assert (
        sendonly.offer_message["encodings"][1]["scaleResolutionDownTo"]["maxWidth"]
        == simulcast_encodings[1]["scaleResolutionDownTo"]["maxWidth"]
    )
    assert (
        sendonly.offer_message["encodings"][2]["scaleResolutionDownTo"]["maxWidth"]
        == simulcast_encodings[2]["scaleResolutionDownTo"]["maxWidth"]
    )

    assert (
        sendonly.offer_message["encodings"][0]["scalabilityMode"]
        == simulcast_encodings[0]["scalabilityMode"]
    )

    assert (
        sendonly.offer_message["encodings"][1]["scalabilityMode"]
        == simulcast_encodings[1]["scalabilityMode"]
    )

    assert (
        sendonly.offer_message["encodings"][2]["scalabilityMode"]
        == simulcast_encodings[2]["scalabilityMode"]
    )

    # "type": "answer" の SDP で Simulcast があるかどうか
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert "a=simulcast:send r0;r1;r2" in sendonly.answer_message["sdp"]

    sendonly_stats = sendonly.get_stats()
    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "video"
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 3

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        # qualityLimitationReason と qualityLimitationDurations がないことはあり得ない
        assert "qualityLimitationReason" in s
        assert "qualityLimitationDurations" in s

        # qualityLimitationReason が "none" ではない場合は正常なテストができないのでテストをスキップする
        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["rid"] == f"r{i}"
        assert s["kind"] == "video"

        # VP8 の場合は scaleResolutionDownTo を指定すると SimulcastEncoderAdapter が無くなる
        # TODO: 念のため他の挙動も確認すること
        if video_codec_type == "VP9":
            assert "SimulcastEncoderAdapter" in s["encoderImplementation"]
        assert encoder_implementation in s["encoderImplementation"]

        assert s["keyFramesEncoded"] > 0
        assert s["bytesSent"] > 500
        assert s["packetsSent"] > 5

        assert s["frameWidth"] == 320
        assert s["frameHeight"] == 176

        assert s["targetBitrate"] >= expect_target_bitrate(
            video_codec_type, s["frameWidth"], s["frameHeight"]
        )

        # Apple Video Toolbox の場合は scalabilityMode がない
        assert "scalabilityMode" not in s

        # targetBitrate が指定したビットレートの 90% 以上、100% 以下に収まることを確認
        print(
            s["rid"],
            video_codec_type,
            s["encoderImplementation"],
            s["targetBitrate"],
            s["frameWidth"],
            s["frameHeight"],
            s["bytesSent"],
            s["packetsSent"],
        )



from client import SoraClient, SoraRole


def test_re_offer_re_answer_sdp(settings):
    recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
    )
    recvonly.connect()

    sendonly1 = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=True,
    )
    sendonly1.connect(fake_audio=True, fake_video=True)
    sendonly1.disconnect()

    sendonly2 = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=True,
    )
    sendonly2.connect(fake_audio=True, fake_video=True)

    sendonly3 = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=True,
    )
    sendonly3.connect(fake_audio=True, fake_video=True)

    sendonly2.disconnect()
    sendonly3.disconnect()

    recvonly.disconnect()



import time

from client import SoraClient, SoraRole


def test_signaling_message(settings):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=True,
    )
    sendonly.connect(fake_audio=True, fake_video=True)

    time.sleep(5)

    sendonly.disconnect()

    assert sendonly.connect_message is not None
    assert sendonly.offer_message is not None
    assert sendonly.answer_message is not None

    assert sendonly.connect_message["role"] == "sendonly"
    assert sendonly.connect_message["channel_id"] == settings.channel_id
    assert sendonly.connect_message["audio"] is True
    assert sendonly.connect_message["video"] is True
    assert sendonly.connect_message["metadata"] == sendonly.metadata


def test_signaling_message_type_connect_forwarding_filter(settings):
    forwarding_filter = {
        "name": "test",
        "priority": 128,
        "action": "block",
        "rules": [
            [
                {
                    "field": "connection_id",
                    "operator": "is_in",
                    "values": ["S8YEN0TSE13JDC2991NG4XZ150"],
                }
            ],
            [
                {"field": "client_id", "operator": "is_in", "values": ["screen-share"]},
                {"field": "kind", "operator": "is_in", "values": ["audio"]},
            ],
        ],
    }

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=True,
        forwarding_filter=forwarding_filter,
    )
    sendonly.connect(fake_audio=True, fake_video=True)

    time.sleep(5)

    sendonly.disconnect()

    assert sendonly.connect_message is not None
    assert sendonly.connect_message["forwarding_filter"] == forwarding_filter


def test_signaling_message_type_connect_forwarding_filters(settings):
    forwarding_filters = [
        {
            "name": "test",
            "priority": 128,
            "action": "block",
            "rules": [
                [
                    {
                        "field": "connection_id",
                        "operator": "is_in",
                        "values": ["S8YEN0TSE13JDC2991NG4XZ150"],
                    }
                ],
                [
                    {"field": "client_id", "operator": "is_in", "values": ["screen-share"]},
                    {"field": "kind", "operator": "is_in", "values": ["audio"]},
                ],
            ],
        },
        {
            "name": "test2",
            "priority": 129,
            "action": "block",
            "rules": [[{"field": "kind", "operator": "is_in", "values": ["audio", "video"]}]],
        },
    ]

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=True,
        forwarding_filters=forwarding_filters,
    )
    sendonly.connect(fake_audio=True, fake_video=True)

    time.sleep(5)

    sendonly.disconnect()

    assert sendonly.connect_message is not None
    assert sendonly.connect_message["forwarding_filters"] == forwarding_filters



import os
import sys
import time

import pytest
from client import SoraClient, SoraDegradationPreference, SoraRole

VIDEO_CODEC_TYPE = "VP8"
VIDEO_BIT_RATE = 300
VIDEO_WIDTH = 960
VIDEO_HEIGHT = 528


@pytest.mark.skipif(
    os.getenv("CI") == "true" and sys.platform == "darwin",
    reason="CI の macOS では性能がでないためスキップする",
)
def test_degradation_preference_maintain_framerate(settings):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=VIDEO_CODEC_TYPE,
        video_bit_rate=VIDEO_BIT_RATE,
        video_width=VIDEO_WIDTH,
        video_height=VIDEO_HEIGHT,
        degradation_preference=SoraDegradationPreference.MAINTAIN_FRAMERATE,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == "video/VP8"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0
    assert outbound_rtp_stats["frameWidth"] <= 640
    assert outbound_rtp_stats["frameHeight"] <= 360
    # ビットレートが 500 kbps 以下
    assert outbound_rtp_stats["targetBitrate"] < VIDEO_BIT_RATE * 1000
    # 20 以上を維持してる
    assert outbound_rtp_stats["framesPerSecond"] > 20

    print(
        outbound_rtp_stats["frameWidth"],
        outbound_rtp_stats["frameHeight"],
        outbound_rtp_stats["framesPerSecond"],
        outbound_rtp_stats["targetBitrate"],
    )


@pytest.mark.skipif(
    os.getenv("CI") == "true" and sys.platform == "darwin",
    reason="CI の macOS では性能がでないためスキップする",
)
def test_degradation_preference_maintain_resolution(settings):
    """
    フレームレートがあまり変わらない
    """
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=VIDEO_CODEC_TYPE,
        video_bit_rate=VIDEO_BIT_RATE,
        video_width=VIDEO_WIDTH,
        video_height=VIDEO_HEIGHT,
        degradation_preference=SoraDegradationPreference.MAINTAIN_RESOLUTION,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{VIDEO_CODEC_TYPE}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0
    # 解像度が維持されてる
    assert outbound_rtp_stats["frameWidth"] == VIDEO_WIDTH
    assert outbound_rtp_stats["frameHeight"] == VIDEO_HEIGHT
    # ビットレートが 500 kbps 以下
    assert outbound_rtp_stats["targetBitrate"] < VIDEO_BIT_RATE * 1000

    print(
        outbound_rtp_stats["frameWidth"],
        outbound_rtp_stats["frameHeight"],
        outbound_rtp_stats["framesPerSecond"],
        outbound_rtp_stats["targetBitrate"],
    )


@pytest.mark.skipif(
    os.getenv("CI") == "true" and sys.platform == "darwin",
    reason="CI の macOS では性能がでないためスキップする",
)
def test_degradation_preference_balanced(settings):
    """
    バランス思った以上に両方悪くなる
    """
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=VIDEO_CODEC_TYPE,
        video_bit_rate=VIDEO_BIT_RATE,
        video_width=VIDEO_WIDTH,
        video_height=VIDEO_HEIGHT,
        degradation_preference=SoraDegradationPreference.BALANCED,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{VIDEO_CODEC_TYPE}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0
    assert outbound_rtp_stats["frameWidth"] <= 640
    assert outbound_rtp_stats["frameHeight"] <= 360
    # ビットレートが 500 kbps 未満
    assert outbound_rtp_stats["targetBitrate"] < VIDEO_BIT_RATE * 1000
    # フレームレートが 30 未満
    assert outbound_rtp_stats["framesPerSecond"] < 30

    print(
        outbound_rtp_stats["frameWidth"],
        outbound_rtp_stats["frameHeight"],
        outbound_rtp_stats["framesPerSecond"],
        outbound_rtp_stats["targetBitrate"],
    )


@pytest.mark.skipif(
    os.getenv("CI") == "true" and sys.platform == "darwin",
    reason="CI の macOS では性能がでないためスキップする",
)
def test_degradation_preference_disabled(settings):
    """
    無効にする
    """
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=VIDEO_CODEC_TYPE,
        video_bit_rate=VIDEO_BIT_RATE,
        video_width=VIDEO_WIDTH,
        video_height=VIDEO_HEIGHT,
        degradation_preference=SoraDegradationPreference.DISABLED,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{VIDEO_CODEC_TYPE}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0
    assert outbound_rtp_stats["frameWidth"] == VIDEO_WIDTH
    assert outbound_rtp_stats["frameHeight"] == VIDEO_HEIGHT
    # ビットレートが 500 kbps 未満
    assert outbound_rtp_stats["targetBitrate"] < VIDEO_BIT_RATE * 1000
    # フレームレートが 30 未満
    assert outbound_rtp_stats["framesPerSecond"] < 30

    print(
        outbound_rtp_stats["frameWidth"],
        outbound_rtp_stats["frameHeight"],
        outbound_rtp_stats["framesPerSecond"],
        outbound_rtp_stats["targetBitrate"],
    )



import time

import pytest
from client import SoraClient, SoraRole


@pytest.mark.skipif(reason="Sora C++ SDK 側の対応が必要")
def test_sendonly_authz_video_true(settings):
    """
    - type: connect で audio: true / video: false で繫ぐ
    - 認証成功時の払い出しで audio: false / video: true を払い出す
    """

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=True,
        video=False,
        jwt_private_claims={
            "audio": True,
            "video": False,
        },
    )
    sendonly.connect(fake_video=False, fake_audio=True)

    time.sleep(5)

    assert sendonly.offer_message is not None
    assert sendonly.offer_message["sdp"] is not None
    assert "VP9" in sendonly.offer_message["sdp"]

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    # 統計で video が見つからないので謎挙動になってる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == "video/VP9"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == "libvpx"
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0


@pytest.mark.parametrize(
    ("video_codec_type", "expected_implementation"),
    [
        # video_codec, expected_implementation
        ("VP8", "libvpx"),
        ("VP9", "libvpx"),
        ("AV1", "libaom"),
    ],
)
def test_sendonly_authz_video_codec_type(settings, video_codec_type, expected_implementation):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        jwt_private_claims={
            "video": True,
            "video_codec_type": video_codec_type,
        },
    )
    sendonly.connect(fake_video=True)

    time.sleep(5)

    assert sendonly.offer_message is not None
    assert sendonly.offer_message["sdp"] is not None
    assert video_codec_type in sendonly.offer_message["sdp"]

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    # 統計で video が見つからないので謎挙動になってる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == expected_implementation
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0



import time

from api import disconnect_connection_api
from client import SoraClient, SoraRole

from sora_sdk import SoraSignalingErrorCode


def test_websocket_signaling_only_disconnect_api(settings):
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=False,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(3)

        if conn.connection_id is None:
            raise ValueError("connection_id is None")
        response = disconnect_connection_api(
            settings.api_url, settings.channel_id, conn.connection_id
        )
        assert response.status_code == 200

        time.sleep(3)

        assert conn.ws_close is True
        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "DISCONNECTED-API"

        assert conn.disconnect_code == SoraSignalingErrorCode.CLOSE_SUCCEEDED
        assert conn.disconnect_reason is not None
        assert "DISCONNECTED-API" in conn.disconnect_reason


def test_websocket_signaling_only_lifetime_expired(settings):
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        jwt_private_claims={
            "audio": False,
            "video": True,
            "connection_lifetime": 3,
        },
        data_channel_signaling=False,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(5)

        assert conn.ws_close is True
        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "LIFETIME-EXPIRED"


def test_websocket_datachannel_signaling_disconnect_api(settings):
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=True,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(3)

        if conn.connection_id is None:
            raise ValueError("connection_id is None")
        response = disconnect_connection_api(
            settings.api_url, settings.channel_id, conn.connection_id
        )
        assert response.status_code == 200

        time.sleep(3)

        assert conn.ws_close is True
        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "DISCONNECTED-API"

        assert conn.disconnect_code == SoraSignalingErrorCode.CLOSE_SUCCEEDED
        assert conn.disconnect_reason is not None
        assert "DISCONNECTED-API" in conn.disconnect_reason


def test_websocket_datachannel_signaling_lifetime_expired(settings):
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        jwt_private_claims={
            "audio": True,
            "video": True,
            "connection_lifetime": 3,
        },
        data_channel_signaling=True,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(5)

        assert conn.ws_close is True
        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "LIFETIME-EXPIRED"


def test_datachannel_only_signaling_disconnect_api(settings):
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=True,
        ignore_disconnect_websocket=True,
    ) as conn:
        time.sleep(3)

        assert conn.switched is True
        assert conn.ignore_disconnect_websocket is True

        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "SELF-CLOSED"

        if conn.connection_id is None:
            raise ValueError("connection_id is None")
        response = disconnect_connection_api(
            settings.api_url, settings.channel_id, conn.connection_id
        )
        assert response.status_code == 200

        time.sleep(3)

        assert conn.close_message is not None
        assert conn.close_message["type"] == "close"
        assert conn.close_message["code"] == 1000
        assert conn.close_message["reason"] == "DISCONNECTED-API"

        assert conn.disconnect_code == SoraSignalingErrorCode.CLOSE_SUCCEEDED
        assert conn.disconnect_reason is not None
        assert "DISCONNECTED-API" in conn.disconnect_reason


def test_datachannel_only_signaling_lifetime_expired(settings):
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        jwt_private_claims={
            "audio": True,
            "video": True,
            "connection_lifetime": 3,
        },
        data_channel_signaling=True,
        ignore_disconnect_websocket=True,
    ) as conn:
        time.sleep(5)

        assert conn.close_message is not None
        assert conn.close_message["type"] == "close"
        assert conn.close_message["code"] == 1000
        assert conn.close_message["reason"] == "LIFETIME-EXPIRED"

        assert conn.disconnect_code == SoraSignalingErrorCode.CLOSE_SUCCEEDED
        assert conn.disconnect_reason is not None
        assert "LIFETIME-EXPIRED" in conn.disconnect_reason



# https://github.com/scikit-build/scikit-build/blob/d06e8724e51885d1d147accfc9f65dd84f1ae7f7/skbuild/cmaker.py より
import configparser
import contextlib
import itertools
import os
import sys
import sysconfig
from pathlib import Path


def get_python_version() -> str:
    """Get version associated with the current python interpreter.

    Returns:
        str: python version string

    Example:
        >>> # xdoc: +IGNORE_WANT
        >>> from skbuild.cmaker import CMaker
        >>> python_version = CMaker.get_python_version()
        >>> print('python_version = {!r}'.format(python_version))
        python_version = '3.7'
    """
    python_version = sysconfig.get_config_var("VERSION")

    if not python_version:
        python_version = sysconfig.get_config_var("py_version_short")

    if not python_version:
        python_version = ".".join(map(str, sys.version_info[:2]))

    assert isinstance(python_version, str)

    return python_version


# NOTE(opadron): The try-excepts raise the cyclomatic complexity, but we
# need them for this function.
def get_python_include_dir(python_version: str):
    """Get include directory associated with the current python
    interpreter.

    Args:
        python_version (str): python version, may be partial.

    Returns:
        PathLike: python include dir

    Example:
        >>> # xdoc: +IGNORE_WANT
        >>> from skbuild.cmaker import CMaker
        >>> python_version = CMaker.get_python_version()
        >>> python_include_dir = CMaker.get_python_include_dir(python_version)
        >>> print('python_include_dir = {!r}'.format(python_include_dir))
        python_include_dir = '.../conda/envs/py37/include/python3.7m'
    """
    # determine python include dir
    python_include_dir: str | None = sysconfig.get_config_var("INCLUDEPY")

    # if Python.h not found (or python_include_dir is None), try to find a
    # suitable include dir
    found_python_h = python_include_dir is not None and os.path.exists(
        os.path.join(python_include_dir, "Python.h")
    )

    if not found_python_h:
        # NOTE(opadron): these possible prefixes must be guarded against
        # AttributeErrors and KeyErrors because they each can throw on
        # different platforms or even different builds on the same platform.
        include_py: str | None = sysconfig.get_config_var("INCLUDEPY")
        include_dir: str | None = sysconfig.get_config_var("INCLUDEDIR")
        include: str | None = None
        plat_include: str | None = None
        python_inc: str | None = None
        python_inc2: str | None = None

        with contextlib.suppress(AttributeError, KeyError):
            include = sysconfig.get_path("include")

        with contextlib.suppress(AttributeError, KeyError):
            plat_include = sysconfig.get_path("platinclude")

        with contextlib.suppress(AttributeError):
            python_inc = sysconfig.get_python_inc()  # type: ignore[attr-defined]

        if include_py is not None:
            include_py = os.path.dirname(include_py)
        if include is not None:
            include = os.path.dirname(include)
        if plat_include is not None:
            plat_include = os.path.dirname(plat_include)
        if python_inc is not None:
            python_inc2 = os.path.join(python_inc, ".".join(map(str, sys.version_info[:2])))

        all_candidate_prefixes = [
            include_py,
            include_dir,
            include,
            plat_include,
            python_inc,
            python_inc2,
        ]
        candidate_prefixes: list[str] = [pre for pre in all_candidate_prefixes if pre]

        candidate_versions: tuple[str, ...] = (python_version,)
        if python_version:
            candidate_versions += ("",)

            pymalloc = None
            with contextlib.suppress(AttributeError):
                pymalloc = bool(sysconfig.get_config_var("WITH_PYMALLOC"))

            if pymalloc:
                candidate_versions += (python_version + "m",)

        candidates = (
            os.path.join(prefix, "".join(("python", ver)))
            for (prefix, ver) in itertools.product(candidate_prefixes, candidate_versions)
        )

        for candidate in candidates:
            if os.path.exists(os.path.join(candidate, "Python.h")):
                # we found an include directory
                python_include_dir = candidate
                break

    # TODO(opadron): what happens if we don't find an include directory?
    #                Throw SKBuildError?

    return python_include_dir


def get_python_library(python_version: str):
    """Get path to the python library associated with the current python
    interpreter.

    Args:
        python_version (str): python version, may be partial.

    Returns:
        PathLike: python_library : python shared library

    Example:
        >>> # xdoc: +IGNORE_WANT
        >>> from skbuild.cmaker import CMaker
        >>> python_version = CMaker.get_python_version()
        >>> python_library = CMaker.get_python_include_dir(python_version)
        >>> print('python_library = {!r}'.format(python_library))
        python_library = '.../conda/envs/py37/include/python3.7m'
    """
    # On Windows, support cross-compiling in the same way as setuptools
    # When cross-compiling, check DIST_EXTRA_CONFIG first
    config_file = os.environ.get("DIST_EXTRA_CONFIG", None)
    if config_file and Path(config_file).is_file():
        cp = configparser.ConfigParser()
        cp.read(config_file)
        result = cp.get("build_ext", "library_dirs", fallback="")
        if result:
            minor = sys.version_info[1]
            return str(Path(result) / f"python3{minor}.lib")

    # This seems to be the simplest way to detect the library path with
    # modern python versions that avoids the complicated construct below.
    # It avoids guessing the library name. Tested with cpython 3.8 and
    # pypy 3.8 on Ubuntu.
    libdir: str | None = sysconfig.get_config_var("LIBDIR")
    ldlibrary: str | None = sysconfig.get_config_var("LDLIBRARY")
    if libdir and ldlibrary and os.path.exists(libdir):
        if sysconfig.get_config_var("MULTIARCH"):
            masd = sysconfig.get_config_var("multiarchsubdir")
            if masd:
                if masd.startswith(os.sep):
                    masd = masd[len(os.sep) :]
                libdir_masd = os.path.join(libdir, masd)
                if os.path.exists(libdir_masd):
                    libdir = libdir_masd
        libpath = os.path.join(libdir, ldlibrary)
        if libpath and os.path.exists(libpath):
            return libpath

    return _guess_python_library(python_version)


def _guess_python_library(python_version: str):
    # determine direct path to libpython
    python_library: str | None = sysconfig.get_config_var("LIBRARY")

    # if static (or nonexistent), try to find a suitable dynamic libpython
    if not python_library or os.path.splitext(python_library)[1][-2:] == ".a":
        candidate_lib_prefixes = ["", "lib"]

        candidate_suffixes = [""]
        candidate_implementations = ["python"]
        if sys.implementation.name == "pypy":
            candidate_implementations[:0] = ["pypy-c", "pypy3-c", "pypy"]
            candidate_suffixes.append("-c")

        candidate_extensions = [".lib", ".so", ".a"]
        # On pypy + MacOS, the variable WITH_DYLD is not set. It would
        # actually be possible to determine the python library there using
        # LDLIBRARY + LIBDIR. As a simple fix, we check if the LDLIBRARY
        # ends with .dylib and add it to the candidate matrix in this case.
        with_ld = sysconfig.get_config_var("WITH_DYLD")
        ld_lib = sysconfig.get_config_var("LDLIBRARY")
        if with_ld or (ld_lib and ld_lib.endswith(".dylib")):
            candidate_extensions.insert(0, ".dylib")

        candidate_versions = [python_version]
        if python_version:
            candidate_versions.append("")
            candidate_versions.insert(0, "".join(python_version.split(".")[:2]))

        abiflags = getattr(sys, "abiflags", "")
        candidate_abiflags = [abiflags]
        if abiflags:
            candidate_abiflags.append("")

        # Ensure the value injected by virtualenv is
        # returned on windows.
        # Because calling `sysconfig.get_config_var('multiarchsubdir')`
        # returns an empty string on Linux, `du_sysconfig` is only used to
        # get the value of `LIBDIR`.
        candidate_libdirs = []
        libdir_a = sysconfig.get_config_var("LIBDIR")
        assert not isinstance(libdir_a, int)
        if libdir_a is None:
            libdest = sysconfig.get_config_var("LIBDEST")
            candidate_libdirs.append(
                os.path.abspath(os.path.join(libdest, "..", "libs") if libdest else "libs")
            )
        libdir_b = sysconfig.get_config_var("LIBDIR")
        for libdir in (libdir_a, libdir_b):
            if libdir is None:
                continue
            if sysconfig.get_config_var("MULTIARCH"):
                masd = sysconfig.get_config_var("multiarchsubdir")
                if masd:
                    if masd.startswith(os.sep):
                        masd = masd[len(os.sep) :]
                    candidate_libdirs.append(os.path.join(libdir, masd))
            candidate_libdirs.append(libdir)

        candidates = (
            os.path.join(libdir, "".join((pre, impl, ver, abi, suf, ext)))
            for (libdir, pre, impl, ext, ver, abi, suf) in itertools.product(
                candidate_libdirs,
                candidate_lib_prefixes,
                candidate_implementations,
                candidate_extensions,
                candidate_versions,
                candidate_abiflags,
                candidate_suffixes,
            )
        )

        for candidate in candidates:
            if os.path.exists(candidate):
                # we found a (likely alternate) libpython
                python_library = candidate
                break

    # Temporary workaround for some libraries (opencv) processing the
    # string output.  Will return None instead of empty string in future
    # versions if the library does not exist.
    if python_library is None:
        return None
    return python_library if python_library and os.path.exists(python_library) else ""



import sys
import time

import pytest
from client import SoraClient, SoraRole
from simulcast import default_video_bit_rate, expect_target_bitrate


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "encoder_implementation",
        "video_width",
        "video_height",
    ),
    [
        # 360p
        ("VP8", "libvpx", 640, 360),
        ("VP9", "libvpx", 640, 360),
        ("AV1", "libaom", 640, 360),
        # 270p
        ("VP8", "libvpx", 480, 270),
        ("VP9", "libvpx", 480, 270),
        ("AV1", "libaom", 480, 270),
    ],
)
def test_authz_simulcast_r2_active_false(
    settings,
    video_codec_type,
    encoder_implementation,
    video_width,
    video_height,
):
    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    simulcast_encodings = [
        {
            "rid": "r0",
            "active": True,
            "scaleResolutionDownBy": 2,
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r1",
            "active": True,
            "scaleResolutionDownBy": 1,
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r2",
            "active": False,
        },
    ]

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        jwt_private_claims={
            "video": True,
            "video_codec_type": video_codec_type,
            "video_bit_rate": video_bit_rate,
            "simulcast": True,
            "simulcast_encodings": simulcast_encodings,
        },
        video_width=video_width,
        video_height=video_height,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    # "type": "offer" の SDP で Simulcast があるかどうか
    assert sendonly.offer_message is not None
    assert sendonly.offer_message["sdp"] is not None
    assert video_codec_type in sendonly.offer_message["sdp"]
    assert "a=simulcast:recv r0;r1;~r2" in sendonly.offer_message["sdp"]

    assert "encodings" in sendonly.offer_message
    assert len(sendonly.offer_message["encodings"]) == 3

    assert sendonly.offer_message["encodings"][0]["rid"] == simulcast_encodings[0]["rid"]
    assert sendonly.offer_message["encodings"][1]["rid"] == simulcast_encodings[1]["rid"]
    assert sendonly.offer_message["encodings"][2]["rid"] == simulcast_encodings[2]["rid"]

    assert sendonly.offer_message["encodings"][0]["active"] == simulcast_encodings[0]["active"]
    assert sendonly.offer_message["encodings"][1]["active"] == simulcast_encodings[1]["active"]
    assert sendonly.offer_message["encodings"][2]["active"] == simulcast_encodings[2]["active"]

    assert (
        sendonly.offer_message["encodings"][0]["scaleResolutionDownBy"]
        == simulcast_encodings[0]["scaleResolutionDownBy"]
    )
    assert (
        sendonly.offer_message["encodings"][1]["scaleResolutionDownBy"]
        == simulcast_encodings[1]["scaleResolutionDownBy"]
    )

    assert (
        sendonly.offer_message["encodings"][0]["scalabilityMode"]
        == simulcast_encodings[0]["scalabilityMode"]
    )

    assert (
        sendonly.offer_message["encodings"][1]["scalabilityMode"]
        == simulcast_encodings[1]["scalabilityMode"]
    )

    # "type": "answer" の SDP で Simulcast があるかどうか
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert "a=simulcast:send r0;r1;~r2" in sendonly.answer_message["sdp"]

    sendonly_stats = sendonly.get_stats()
    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s
        for s in sendonly_stats
        if s.get("type") == "outbound-rtp" and s.get("kind") == "video" and s.get("active") is True
    ]

    # active は 2 本だけ
    assert len(outbound_rtp_stats) == 2

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert s["rid"] == f"r{i}"
        assert s["kind"] == "video"

        assert encoder_implementation in s.get("encoderImplementation")
        if video_codec_type in ["AV1"]:
            assert "SimulcastEncoderAdapter" in s.get("encoderImplementation")

        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["keyFramesEncoded"] > 0
        assert s["bytesSent"] > 500
        assert s["packetsSent"] > 5

        # assert s["frameWidth"] >= video_width
        # assert s["frameHeight"] >= video_height

        assert s["targetBitrate"] >= expect_target_bitrate(
            video_codec_type, s["frameWidth"], s["frameHeight"]
        )

        assert "scalabilityMode" in s
        assert s["scalabilityMode"] == "L1T1"

        print(
            s["rid"],
            video_codec_type,
            s["encoderImplementation"],
            s["scalabilityMode"],
            video_bit_rate * 1000,
            s["targetBitrate"],
            s["frameWidth"],
            s["frameHeight"],
            s["bytesSent"],
            s["packetsSent"],
        )


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "encoder_implementation",
        "video_width",
        "video_height",
    ),
    [
        # 180p
        ("VP8", "libvpx", 320, 180),
        ("VP9", "libvpx", 320, 180),
        ("AV1", "libaom", 320, 180),
        # 135p
        ("VP8", "libvpx", 240, 135),
        ("VP9", "libvpx", 240, 135),
        ("AV1", "libaom", 240, 135),
    ],
)
def test_authz_simulcast_r2_and_r1_active_false(
    settings,
    video_codec_type,
    encoder_implementation,
    video_width,
    video_height,
):
    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    simulcast_encodings = [
        {
            "rid": "r0",
            "active": True,
            "scaleResolutionDownBy": 1,
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r1",
            "active": False,
        },
        {
            "rid": "r2",
            "active": False,
        },
    ]

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        jwt_private_claims={
            "video": True,
            "video_codec_type": video_codec_type,
            "video_bit_rate": video_bit_rate,
            "simulcast": True,
            "simulcast_encodings": simulcast_encodings,
        },
        video_width=video_width,
        video_height=video_height,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    # "type": "offer" の SDP で Simulcast があるかどうか
    assert sendonly.offer_message is not None
    assert sendonly.offer_message["sdp"] is not None
    assert video_codec_type in sendonly.offer_message["sdp"]
    assert "a=simulcast:recv r0;~r1;~r2" in sendonly.offer_message["sdp"]

    assert "encodings" in sendonly.offer_message
    assert len(sendonly.offer_message["encodings"]) == 3

    assert sendonly.offer_message["encodings"][0]["rid"] == simulcast_encodings[0]["rid"]
    assert sendonly.offer_message["encodings"][1]["rid"] == simulcast_encodings[1]["rid"]
    assert sendonly.offer_message["encodings"][2]["rid"] == simulcast_encodings[2]["rid"]

    assert sendonly.offer_message["encodings"][0]["active"] == simulcast_encodings[0]["active"]
    assert sendonly.offer_message["encodings"][1]["active"] == simulcast_encodings[1]["active"]
    assert sendonly.offer_message["encodings"][2]["active"] == simulcast_encodings[2]["active"]

    assert (
        sendonly.offer_message["encodings"][0]["scaleResolutionDownBy"]
        == simulcast_encodings[0]["scaleResolutionDownBy"]
    )

    assert (
        sendonly.offer_message["encodings"][0]["scalabilityMode"]
        == simulcast_encodings[0]["scalabilityMode"]
    )

    # "type": "answer" の SDP で Simulcast があるかどうか
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert "a=simulcast:send r0;~r1;~r2" in sendonly.answer_message["sdp"]

    sendonly_stats = sendonly.get_stats()
    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s
        for s in sendonly_stats
        if s.get("type") == "outbound-rtp" and s.get("kind") == "video" and s.get("active") is True
    ]

    assert len(outbound_rtp_stats) == 1

    s = outbound_rtp_stats[0]

    assert s["rid"] == "r0"
    assert s["kind"] == "video"

    assert encoder_implementation in s["encoderImplementation"]
    print(s["encoderImplementation"])

    if s["qualityLimitationReason"] != "none":
        pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

    assert s["keyFramesEncoded"] > 0
    assert s["bytesSent"] > 500
    assert s["packetsSent"] > 5

    assert s["targetBitrate"] >= expect_target_bitrate(
        video_codec_type, s["frameWidth"], s["frameHeight"]
    )

    assert "scalabilityMode" in s
    assert s["scalabilityMode"] == "L1T1"

    print(
        s["rid"],
        video_codec_type,
        s["encoderImplementation"],
        s["scalabilityMode"],
        video_bit_rate * 1000,
        s["targetBitrate"],
        s["frameWidth"],
        s["frameHeight"],
        s["bytesSent"],
        s["packetsSent"],
    )


@pytest.mark.skipif(sys.platform == "darwin", reason="Apple では SW コーデックは動作させない")
@pytest.mark.parametrize(
    (
        "video_codec_type",
        "encoder_implementation",
        "video_bit_rate",
        "video_width",
        "video_height",
    ),
    [
        # どうやら scaleResolutionDownTo を指定すると規定されたテーブルのビットレートでは足りない模様
        ("VP8", "libvpx", 1200 * 3, 960, 540),
        ("VP9", "libvpx", 879 * 3, 960, 540),
        ("AV1", "libaom", 879 * 3, 960, 540),
    ],
)
def test_authz_simulcast_scale_resolution_down_to(
    settings,
    video_codec_type,
    encoder_implementation,
    video_bit_rate,
    video_width,
    video_height,
):
    simulcast_encodings = [
        {
            "rid": "r0",
            "active": True,
            "scaleResolutionDownTo": {"maxWidth": 640, "maxHeight": 360},
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r1",
            "active": True,
            "scaleResolutionDownTo": {"maxWidth": 640, "maxHeight": 360},
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r2",
            "active": True,
            "scaleResolutionDownTo": {"maxWidth": 640, "maxHeight": 360},
            "scalabilityMode": "L1T1",
        },
    ]

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        jwt_private_claims={
            "video": True,
            "video_codec_type": video_codec_type,
            "video_bit_rate": video_bit_rate,
            "simulcast": True,
            "simulcast_encodings": simulcast_encodings,
        },
        video_width=video_width,
        video_height=video_height,
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    # "type": "offer" の SDP で Simulcast があるかどうか
    assert sendonly.offer_message is not None
    assert sendonly.offer_message["sdp"] is not None
    assert video_codec_type in sendonly.offer_message["sdp"]
    assert "a=simulcast:recv r0;r1;r2" in sendonly.offer_message["sdp"]

    assert "encodings" in sendonly.offer_message
    assert len(sendonly.offer_message["encodings"]) == 3

    assert sendonly.offer_message["encodings"][0]["rid"] == simulcast_encodings[0]["rid"]
    assert sendonly.offer_message["encodings"][1]["rid"] == simulcast_encodings[1]["rid"]
    assert sendonly.offer_message["encodings"][2]["rid"] == simulcast_encodings[2]["rid"]

    assert sendonly.offer_message["encodings"][0]["active"] == simulcast_encodings[0]["active"]
    assert sendonly.offer_message["encodings"][1]["active"] == simulcast_encodings[1]["active"]
    assert sendonly.offer_message["encodings"][2]["active"] == simulcast_encodings[2]["active"]

    assert (
        sendonly.offer_message["encodings"][0]["scaleResolutionDownTo"]["maxWidth"]
        == simulcast_encodings[0]["scaleResolutionDownTo"]["maxWidth"]
    )
    assert (
        sendonly.offer_message["encodings"][1]["scaleResolutionDownTo"]["maxWidth"]
        == simulcast_encodings[1]["scaleResolutionDownTo"]["maxWidth"]
    )
    assert (
        sendonly.offer_message["encodings"][2]["scaleResolutionDownTo"]["maxWidth"]
        == simulcast_encodings[2]["scaleResolutionDownTo"]["maxWidth"]
    )

    assert (
        sendonly.offer_message["encodings"][0]["scalabilityMode"]
        == simulcast_encodings[0]["scalabilityMode"]
    )

    assert (
        sendonly.offer_message["encodings"][1]["scalabilityMode"]
        == simulcast_encodings[1]["scalabilityMode"]
    )

    assert (
        sendonly.offer_message["encodings"][2]["scalabilityMode"]
        == simulcast_encodings[2]["scalabilityMode"]
    )

    # "type": "answer" の SDP で Simulcast があるかどうか
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert "a=simulcast:send r0;r1;r2" in sendonly.answer_message["sdp"]

    sendonly_stats = sendonly.get_stats()
    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "video"
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 3

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert s["rid"] == f"r{i}"
        assert s["kind"] == "video"

        # VP8 の場合は scaleResolutionDownTo を指定すると SimulcastEncoderAdapter が無くなる
        # TODO: 念のため他の挙動も確認すること
        if video_codec_type == "VP9":
            assert "SimulcastEncoderAdapter" in s["encoderImplementation"]
        assert encoder_implementation in s["encoderImplementation"]

        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["keyFramesEncoded"] > 0
        assert s["bytesSent"] > 500
        assert s["packetsSent"] > 5

        assert s["frameWidth"] == 640
        assert s["frameHeight"] == 352

        assert s["targetBitrate"] >= expect_target_bitrate(
            video_codec_type, s["frameWidth"], s["frameHeight"]
        )

        scalability_mode = None
        if "scalabilityMode" in s:
            assert s["scalabilityMode"] == "L1T1"
            scalability_mode = s["scalabilityMode"]

        # targetBitrate が指定したビットレートの 90% 以上、100% 以下に収まることを確認
        expected_bitrate = video_bit_rate * 1000
        print(
            s["rid"],
            video_codec_type,
            s["encoderImplementation"],
            scalability_mode,
            expected_bitrate,
            s["targetBitrate"],
            s["frameWidth"],
            s["frameHeight"],
            s["bytesSent"],
            s["packetsSent"],
        )



import os
import time

import pytest
from api import request_key_frame_api
from client import (
    SoraClient,
    SoraRole,
    codec_type_string_to_codec_type,
    get_video_codec_capability,
)
from simulcast import default_video_bit_rate, expect_target_bitrate

from sora_sdk import (
    SoraVideoCodecImplementation,
    SoraVideoCodecPreference,
    SoraVideoCodecType,
)

pytestmark = pytest.mark.skipif(
    os.environ.get("NVIDIA_VIDEO_CODEC") is None, reason="NVIDIA Video Codec でのみ実行する"
)


def test_nvidia_video_codec_sdk_available():
    capability = get_video_codec_capability()

    nvidia_video_codec_sdk_available = False
    for e in capability.engines:
        if e.name == SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK:
            nvidia_video_codec_sdk_available = True

    assert nvidia_video_codec_sdk_available is True

    for e in capability.engines:
        if e.name == SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK:
            # 対応コーデックは 5 種類
            assert len(e.codecs) == 5

            for c in e.codecs:
                match c.type:
                    case SoraVideoCodecType.VP8:
                        assert c.decoder is True
                        assert c.encoder is False
                    case SoraVideoCodecType.VP9:
                        assert c.decoder is True
                        assert c.encoder is False
                    case SoraVideoCodecType.AV1:
                        assert c.decoder is True
                        assert c.encoder is True
                    case SoraVideoCodecType.H264:
                        assert c.decoder is True
                        assert c.encoder is True
                    case SoraVideoCodecType.H265:
                        assert c.decoder is True
                        assert c.encoder is True
                    case _:
                        pytest.fail(f"未実装の codec_type: {c.type}")


@pytest.mark.parametrize(
    "video_codec_type",
    [
        "AV1",
        "H264",
        "H265",
    ],
)
def test_intel_vpl_key_frame_request(settings, video_codec_type):
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=codec_type_string_to_codec_type(video_codec_type),
                    encoder=SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK,
                ),
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    time.sleep(3)

    assert sendonly.connection_id is not None

    # キーフレーム要求 API を 3 秒間隔で 3 回呼び出す
    api_count = 3
    for _ in range(api_count):
        response = request_key_frame_api(
            settings.api_url, sendonly.channel_id, sendonly.connection_id
        )
        assert response.status_code == 200
        time.sleep(3)

    # 統計を取得する
    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")

    # 3 回以上
    assert outbound_rtp_stats["keyFramesEncoded"] > api_count
    assert outbound_rtp_stats["pliCount"] >= api_count
    print("keyFramesEncoded:", outbound_rtp_stats["keyFramesEncoded"])
    print("pliCount:", outbound_rtp_stats["pliCount"])

    # PLI カウントの 50% 以上がキーフレームとしてエンコードされることを確認
    assert outbound_rtp_stats["keyFramesEncoded"] >= outbound_rtp_stats["pliCount"] * 0.7
    print(
        "keyFramesEncoded >= pliCount * 0.7:",
        outbound_rtp_stats["keyFramesEncoded"] >= outbound_rtp_stats["pliCount"] * 0.7,
    )


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "expected_implementation",
    ),
    [
        ("AV1", "NvCodec"),
        ("H264", "NvCodec"),
        ("H265", "NvCodec"),
    ],
)
def test_nvidia_codec_sdk_sendonly_recvonly(settings, video_codec_type, expected_implementation):
    # if not is_codec_supported(
    #     video_codec_type, SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK
    # ):
    #     pytest.skip(f"このチップでは {video_codec_type} がサポートされていません")

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=codec_type_string_to_codec_type(video_codec_type),
                    encoder=SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK,
                ),
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=codec_type_string_to_codec_type(video_codec_type),
                    decoder=SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK,
                ),
            ]
        ),
    )
    recvonly.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    recvonly_stats = recvonly.get_stats()

    sendonly.disconnect()
    recvonly.disconnect()

    # offer の sdp に video_codec_type が含まれているかどうかを確認している
    assert sendonly.offer_message is not None
    assert "sdp" in sendonly.offer_message
    assert video_codec_type in sendonly.offer_message["sdp"]

    # answer の sdp に video_codec_type が含まれているかどうかを確認している
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert video_codec_type in sendonly.answer_message["sdp"]

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    # H.264/H.265 が採用されているかどうか確認する
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == expected_implementation
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # codec が無かったら StopIteration 例外が上がる
    recvonly_codec_stats = next(s for s in recvonly_stats if s.get("type") == "codec")
    # H.264/H.265 が採用されているかどうか確認する
    assert recvonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # inbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(s for s in recvonly_stats if s.get("type") == "inbound-rtp")
    assert inbound_rtp_stats["decoderImplementation"] == expected_implementation
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "expected_implementation",
        "video_width",
        "video_height",
        "simulcast_count",
    ),
    # FIXME: H.265 では、本数が 2 本 や 1 本の場合、エラーになるのでコメントアウトしている
    # FIXME: AV1 では、解像度が一定数より低くなる場合、エラーになるのでコメントアウトしている
    [
        # 1080p
        ("AV1", "NvCodec", 1920, 1080, 3),
        ("H264", "NvCodec", 1920, 1080, 3),
        ("H265", "NvCodec", 1920, 1080, 3),
        # 720p
        ("AV1", "NvCodec", 1280, 720, 3),
        ("H264", "NvCodec", 1280, 720, 3),
        ("H265", "NvCodec", 1280, 720, 3),
        # 540p
        ("AV1", "NvCodec", 960, 540, 3),
        ("H264", "NvCodec", 960, 540, 3),
        ("H265", "NvCodec", 960, 540, 3),
        # 360p
        # ("AV1", "NvCodec", 640, 360, 2),
        # ("H264", "NvCodec", 640, 360, 2),
        # ("H265", "NvCodec", 640, 360, 2),
        # 270p
        # ("AV1", "NvCodec", 480, 270, 2),
        # ("H264", "NvCodec", 480, 270, 2),
        # ("H265", "NvCodec", 480, 270, 2),
        # 180p
        # ("H264", "NvCodec", 320, 180, 1),
        # ("H265", "NvCodec", 320, 180, 1),
        # 135p
        # ("H265", "NvCodec", 240, 135, 1),
    ],
)
def test_nvidia_video_codec_sdk_simulcast(
    settings,
    video_codec_type,
    expected_implementation,
    video_width,
    video_height,
    simulcast_count,
):
    # if not is_codec_supported(
    #     video_codec_type, SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK
    # ):
    #     pytest.skip(f"このチップでは {video_codec_type} がサポートされていません")

    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        simulcast=True,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_bit_rate=video_bit_rate,
        video_width=video_width,
        video_height=video_height,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=codec_type_string_to_codec_type(video_codec_type),
                    encoder=SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK,
                ),
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # "type": "answer" の SDP で Simulcast があるかどうか
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert "a=simulcast:send r0;r1;r2" in sendonly.answer_message["sdp"]

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "video"
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 3

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert "qualityLimitationReason" in s
        assert "qualityLimitationDurations" in s

        # qualityLimitationReason が none で無い場合は安定したテストができない
        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["rid"] == f"r{i}"
        # simulcast_count が 2 の場合、rid r2 の bytesSent/packetsSent は 0 or 1 になる
        # simulcast_count が 1 の場合、rid r2 と r1 の bytesSent/packetsSent は 0 or 1 になる
        if i < simulcast_count:
            # 1 本になると simulcastEncodingAdapter がなくなる
            # if simulcast_count > 1:
            #     assert "SimulcastEncoderAdapter" in s["encoderImplementation"]
            assert expected_implementation in s["encoderImplementation"]

            assert s["bytesSent"] > 500
            assert s["packetsSent"] > 5

            assert s["targetBitrate"] >= expect_target_bitrate(
                video_codec_type, s["frameWidth"], s["frameHeight"]
            )

            print(
                s["rid"],
                video_codec_type,
                expected_implementation,
                video_bit_rate * 1000,
                s["targetBitrate"],
                s["frameWidth"],
                s["frameHeight"],
                s["bytesSent"],
                s["packetsSent"],
            )
        else:
            # 本来は 0 なのだが、simulcast_count が 1 の場合、
            # packetSent が 0 ではなく 1 や 2 になる場合がある
            # byteSent は 0
            print(
                s["rid"],
                video_codec_type,
                s["bytesSent"],
                s["packetsSent"],
            )
            assert s["bytesSent"] == 0
            assert s["packetsSent"] <= 2


# VP8 / VP9 は HWA Decoder のみ搭載している
# https://developer.nvidia.com/video-encode-and-decode-gpu-support-matrix-new
@pytest.mark.parametrize(
    ("video_codec_type", "expected_implementation"),
    [
        ("VP8", "NvCodec"),
        ("VP9", "NvCodec"),
    ],
)
def test_nvidia_video_codec_sdk_decoding_only(settings, video_codec_type, expected_implementation):
    """
    NVIDIA Video Codec SDK VP8/VP9 はデコーダーは利用できるので、そのテスト
    """
    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=codec_type_string_to_codec_type(video_codec_type),
                    encoder=SoraVideoCodecImplementation.INTERNAL,
                ),
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    recvonly = SoraClient(
        settings,
        SoraRole.RECVONLY,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=codec_type_string_to_codec_type(video_codec_type),
                    decoder=SoraVideoCodecImplementation.NVIDIA_VIDEO_CODEC_SDK,
                ),
            ]
        ),
    )
    recvonly.connect()

    time.sleep(5)

    sendonly_stats = sendonly.get_stats()
    recvonly_stats = recvonly.get_stats()

    sendonly.disconnect()
    recvonly.disconnect()

    # offer の sdp に video_codec_type が含まれているかどうかを確認している
    assert sendonly.offer_message is not None
    assert "sdp" in sendonly.offer_message
    assert video_codec_type in sendonly.offer_message["sdp"]

    # answer の sdp に video_codec_type が含まれているかどうかを確認している
    assert sendonly.answer_message is not None
    assert "sdp" in sendonly.answer_message
    assert video_codec_type in sendonly.answer_message["sdp"]

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    # VP9 が採用されているかどうか確認する
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # outbound-rtp が無かったら StopIteration 例外が上がる
    outbound_rtp_stats = next(s for s in sendonly_stats if s.get("type") == "outbound-rtp")
    assert outbound_rtp_stats["encoderImplementation"] == "libvpx"
    assert outbound_rtp_stats["bytesSent"] > 0
    assert outbound_rtp_stats["packetsSent"] > 0

    # codec が無かったら StopIteration 例外が上がる
    recvonly_codec_stats = next(s for s in recvonly_stats if s.get("type") == "codec")
    # VP9 が採用されているかどうか確認する
    assert recvonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # inbound-rtp が無かったら StopIteration 例外が上がる
    inbound_rtp_stats = next(s for s in recvonly_stats if s.get("type") == "inbound-rtp")
    assert inbound_rtp_stats["decoderImplementation"] == expected_implementation
    assert inbound_rtp_stats["bytesReceived"] > 0
    assert inbound_rtp_stats["packetsReceived"] > 0



import os
import sys
import sysconfig

from setuptools import setup
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, BASE_DIR)


from buildbase import PlatformTarget, cd, get_build_platform  # noqa: E402


def run_setup(build_platform, target_platform):
    with open(os.path.join(BASE_DIR, "VERSION"), "r") as f:
        version = f.read().strip()

    build_profile = os.getenv("BUILD_PROFILE")
    if build_profile == "debug":
        version += "+debug"

    plat = None
    additional_files = []
    if target_platform.os == "jetson":
        plat = "manylinux_2_17_aarch64.manylinux2014_aarch64"
    elif target_platform.os == "ubuntu" and target_platform.arch == "armv8":
        if target_platform.osver == "22.04":
            plat = "manylinux_2_31_aarch64"
        if target_platform.osver == "24.04":
            plat = "manylinux_2_35_aarch64"
    elif target_platform.os == "ubuntu" and target_platform.arch == "x86_64":
        if target_platform.osver == "22.04":
            plat = "manylinux_2_31_x86_64"
        if target_platform.osver == "24.04":
            plat = "manylinux_2_35_x86_64"
    elif target_platform.os == "raspberry-pi-os":
        plat = "manylinux_2_35_aarch64"
        additional_files += ["libcamerac.so"]

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False

        def get_tag(self):
            _, _, plat2 = super().get_tag()
            impl = "cp" + sysconfig.get_config_var("py_version_nodot")
            return impl, impl, plat if plat is not None else plat2

    setup(
        version=version,
        url="https://github.com/shiguredo/sora-python-sdk",
        packages=["sora_sdk"],
        package_dir={"": "src"},
        package_data={
            "sora_sdk": ["sora_sdk_ext.*", *additional_files],
        },
        include_package_data=True,
        cmdclass={
            "bdist_wheel": bdist_wheel,
        },
    )


def main():
    build_platform = get_build_platform()

    target = os.getenv("SORA_SDK_TARGET")
    if target is None:
        target_platform = build_platform
    elif target == "ubuntu-22.04_armv8":
        target_platform = PlatformTarget("ubuntu", "22.04", "armv8")
    elif target == "ubuntu-24.04_armv8":
        target_platform = PlatformTarget("ubuntu", "24.04", "armv8")
    elif target == "ubuntu-22.04_armv8_jetson":
        target_platform = PlatformTarget("jetson", None, "armv8", "ubuntu-22.04")
    elif target == "raspberry-pi-os_armv8":
        target_platform = PlatformTarget("raspberry-pi-os", None, "armv8")
    else:
        raise Exception(f"Unknown target {target}")

    with cd(BASE_DIR):
        run_setup(build_platform, target_platform)


if __name__ == "__main__":
    main()



import os
import time

import pytest
from client import SoraClient, SoraRole
from simulcast import default_video_bit_rate, expect_target_bitrate

from sora_sdk import (
    SoraVideoCodecImplementation,
    SoraVideoCodecPreference,
    SoraVideoCodecType,
)

pytestmark = pytest.mark.skipif(
    os.environ.get("OPENH264_PATH") is None,
    reason="OpenH264 のときだけ実行する",
)


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "expected_implementation",
        "video_width",
        "video_height",
        "simulcast_count",
    ),
    [
        # 1080p
        ("H264", "OpenH264", 1920, 1080, 3),
        # 720p
        ("H264", "OpenH264", 1280, 720, 3),
        # 540p
        ("H264", "OpenH264", 960, 540, 3),
    ],
)
def test_openh264_simulcast(
    settings,
    video_codec_type,
    expected_implementation,
    video_width,
    video_height,
    simulcast_count,
):
    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        simulcast=True,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_bit_rate=video_bit_rate,
        video_width=video_width,
        video_height=video_height,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.H264,
                    encoder=SoraVideoCodecImplementation.CISCO_OPENH264,
                )
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s for s in sendonly_stats if s.get("type") == "outbound-rtp" and s.get("kind") == "video"
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 3

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert "qualityLimitationReason" in s
        assert "qualityLimitationDurations" in s

        # qualityLimitationReason が none で無い場合は安定したテストができない
        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["rid"] == f"r{i}"
        # simulcast_count が 2 の場合、rid r2 の bytesSent/packetsSent は 0 or 1 になる
        # simulcast_count が 1 の場合、rid r2 と r1 の bytesSent/packetsSent は 0 or 1 になる
        if i < simulcast_count:
            # 1 本になると simulcastEncodingAdapter がなくなる
            if simulcast_count > 1:
                assert "SimulcastEncoderAdapter" in s["encoderImplementation"]
            assert expected_implementation in s["encoderImplementation"]

            assert s["bytesSent"] > 500
            assert s["packetsSent"] > 5

            assert s["targetBitrate"] >= expect_target_bitrate(
                video_codec_type, s["frameWidth"], s["frameHeight"]
            )

            scalability_mode = None
            if "scalabilityMode" in s:
                assert s["scalabilityMode"] == "L1T1"
                scalability_mode = s["scalabilityMode"]

            print(
                s["rid"],
                video_codec_type,
                expected_implementation,
                scalability_mode,
                video_bit_rate * 1000,
                s["targetBitrate"],
                s["frameWidth"],
                s["frameHeight"],
                s["bytesSent"],
                s["packetsSent"],
            )
        else:
            # 本来は 0 なのだが、simulcast_count が 1 の場合、
            # packetSent が 0 ではなく 1 や 2 になる場合がある
            # byteSent は 0
            assert s["bytesSent"] == 0
            assert s["packetsSent"] <= 2
            print(
                s["rid"],
                video_codec_type,
                s["bytesSent"],
                s["packetsSent"],
            )


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "expected_implementation",
        "video_width",
        "video_height",
    ),
    [
        # 360p
        ("H264", "OpenH264", 640, 360),
        # 270p
        ("H264", "OpenH264", 480, 270),
    ],
)
def test_openh264_authz_simulcast_r2_active_false(
    settings,
    video_codec_type,
    expected_implementation,
    video_width,
    video_height,
):
    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    simulcast_encodings = [
        {
            "rid": "r0",
            "active": True,
            "scaleResolutionDownBy": 2,
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r1",
            "active": True,
            "scaleResolutionDownBy": 1,
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r2",
            "active": False,
        },
    ]

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        simulcast=True,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_bit_rate=video_bit_rate,
        jwt_private_claims={
            "video": True,
            "video_codec_type": video_codec_type,
            "video_bit_rate": video_bit_rate,
            "simulcast": True,
            "simulcast_encodings": simulcast_encodings,
        },
        video_width=video_width,
        video_height=video_height,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.H264,
                    encoder=SoraVideoCodecImplementation.CISCO_OPENH264,
                )
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s
        for s in sendonly_stats
        if s.get("type") == "outbound-rtp" and s.get("kind") == "video" and s.get("active") is True
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 2

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert "qualityLimitationReason" in s
        assert "qualityLimitationDurations" in s

        # qualityLimitationReason が none で無い場合は安定したテストができない
        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["rid"] == f"r{i}"
        print(s.get("encoderImplementation"))
        assert expected_implementation in s["encoderImplementation"]

        assert s["bytesSent"] > 500
        assert s["packetsSent"] > 5

        assert s["targetBitrate"] >= expect_target_bitrate(
            video_codec_type, s["frameWidth"], s["frameHeight"]
        )

        assert s["scalabilityMode"] == "L1T1"

        print(
            s["rid"],
            video_codec_type,
            expected_implementation,
            video_bit_rate * 1000,
            s["targetBitrate"],
            s["frameWidth"],
            s["frameHeight"],
            s["bytesSent"],
            s["packetsSent"],
        )


@pytest.mark.parametrize(
    (
        "video_codec_type",
        "expected_implementation",
        "video_width",
        "video_height",
    ),
    [
        # 180p
        ("H264", "OpenH264", 320, 180),
    ],
)
def test_openh264_authz_simulcast_r2_and_r1_active_false(
    settings,
    video_codec_type,
    expected_implementation,
    video_width,
    video_height,
):
    video_bit_rate = default_video_bit_rate(video_codec_type, video_width, video_height)

    simulcast_encodings = [
        {
            "rid": "r0",
            "active": True,
            "scaleResolutionDownBy": 1,
            "scalabilityMode": "L1T1",
        },
        {
            "rid": "r1",
            "active": False,
        },
        {
            "rid": "r2",
            "active": False,
        },
    ]

    sendonly = SoraClient(
        settings,
        SoraRole.SENDONLY,
        simulcast=True,
        audio=False,
        video=True,
        video_codec_type=video_codec_type,
        video_bit_rate=video_bit_rate,
        jwt_private_claims={
            "video": True,
            "video_codec_type": video_codec_type,
            "video_bit_rate": video_bit_rate,
            "simulcast": True,
            "simulcast_encodings": simulcast_encodings,
        },
        video_width=video_width,
        video_height=video_height,
        video_codec_preference=SoraVideoCodecPreference(
            codecs=[
                SoraVideoCodecPreference.Codec(
                    type=SoraVideoCodecType.H264,
                    encoder=SoraVideoCodecImplementation.CISCO_OPENH264,
                )
            ]
        ),
    )
    sendonly.connect(fake_video=True)

    time.sleep(10)

    sendonly_stats = sendonly.get_stats()

    sendonly.disconnect()

    # codec が無かったら StopIteration 例外が上がる
    sendonly_codec_stats = next(s for s in sendonly_stats if s.get("type") == "codec")
    assert sendonly_codec_stats["mimeType"] == f"video/{video_codec_type}"

    # 複数の outbound-rtp 統計情報を取得
    outbound_rtp_stats = [
        s
        for s in sendonly_stats
        if s.get("type") == "outbound-rtp" and s.get("kind") == "video" and s.get("active") is True
    ]
    # simulcast_count に関係なく統計情報はかならず 3 本出力される
    # これは SDP で rid で ~r0 とかやる減るはず
    assert len(outbound_rtp_stats) == 1

    # rid でソート
    sorted_stats = sorted(outbound_rtp_stats, key=lambda x: x.get("rid", ""))

    for i, s in enumerate(sorted_stats):
        assert "qualityLimitationReason" in s
        assert "qualityLimitationDurations" in s

        # qualityLimitationReason が none で無い場合は安定したテストができない
        if s["qualityLimitationReason"] != "none":
            pytest.skip(f"qualityLimitationReason: {s['qualityLimitationReason']}")

        assert s["rid"] == f"r{i}"
        print(s.get("encoderImplementation"))
        assert expected_implementation in s["encoderImplementation"]

        assert s["bytesSent"] > 500
        assert s["packetsSent"] > 5

        assert s["targetBitrate"] >= expect_target_bitrate(
            video_codec_type, s["frameWidth"], s["frameHeight"]
        )

        assert s["scalabilityMode"] == "L1T1"

        print(
            s["rid"],
            video_codec_type,
            expected_implementation,
            video_bit_rate * 1000,
            s["targetBitrate"],
            s["frameWidth"],
            s["frameHeight"],
            s["bytesSent"],
            s["packetsSent"],
        )



import os
import sora_sdk


def test_version():
    """sora_sdk.__version__ が取得できることを確認"""
    assert hasattr(sora_sdk, "__version__")
    assert isinstance(sora_sdk.__version__, str)
    assert sora_sdk.__version__ != "unknown"

    # VERSION ファイルの内容と一致することを確認
    version_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), "VERSION")
    with open(version_file, "r") as f:
        expected_version = f.read().strip()

    assert sora_sdk.__version__ == expected_version
    print(f"sora_sdk.__version__ = {sora_sdk.__version__}")



import time

from client import SoraClient, SoraRole

from sora_sdk import SoraSignalingErrorCode


def test_websocket_signaling_only_disconnect(settings):
    """
    - WebSocket シグナリングのみ
    """

    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=False,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(3)

        conn.disconnect()

        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "TYPE-DISCONNECT"

        assert conn.disconnect_code == SoraSignalingErrorCode.CLOSE_SUCCEEDED
        assert (
            conn.disconnect_reason == "Succeeded to close WebSocket (DC signaling is not enabled)"
        )


def test_hybrid_signaling_disconnect(settings):
    """
    - WebSocket シグナリングと DataChannel シグナリング
    """
    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=True,
        ignore_disconnect_websocket=False,
    ) as conn:
        time.sleep(3)

        assert conn.switched is True

        conn.disconnect()

        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "TYPE-DISCONNECT"

        assert conn.disconnect_code == SoraSignalingErrorCode.CLOSE_SUCCEEDED
        assert conn.disconnect_reason == "Succeeded to close Websocket (DC signaling is enabled)"


def test_datachannel_only_type_disconnect(settings):
    """
    - DataChannel シグナリングのみ
    """

    with SoraClient(
        settings,
        SoraRole.RECVONLY,
        audio=True,
        video=True,
        data_channel_signaling=True,
        ignore_disconnect_websocket=True,
    ) as conn:
        time.sleep(5)

        assert conn.switched is True
        assert conn.ws_close_code == 1000
        assert conn.ws_close_reason == "SELF-CLOSED"

        conn.disconnect()

        assert conn.disconnect_code == SoraSignalingErrorCode.CLOSE_SUCCEEDED
        assert conn.disconnect_reason == "Succeeded to close DataChannel"



from .sora_sdk_ext import *  # noqa: F401,F403

# インストールされたパッケージの場合は importlib.metadata から取得
# 開発環境の場合は VERSION ファイルから取得
try:
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version("sora_sdk")
    except PackageNotFoundError:
        # パッケージがインストールされていない場合（開発環境）
        import os

        _version_file = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "VERSION"
        )
        if os.path.exists(_version_file):
            with open(_version_file, "r") as f:
                __version__ = f.read().strip()
        else:
            __version__ = "unknown"
except ImportError:
    # Python 3.8 以前の互換性のため
    __version__ = "unknown"

"""
sink はそれぞれ track が必要で参照を保持する必要がある
しかしながら、 sink の C++ 側で shared_ptr として track を持つと、
リファレンスカウンタが正しく処理されず終了時にリークしてしまう。
そのため Python で Wrapper を作り、その中で保持することとした。
"""


class SoraAudioSink(SoraAudioSinkImpl):
    def __init__(self, track, output_frequency, output_channels):
        super().__init__(track, output_frequency, output_channels)
        self.__track = track

    def __del__(self):
        super().__del__()
        del self.__track


class SoraAudioStreamSink(SoraAudioStreamSinkImpl):
    def __init__(self, track, output_frequency, output_channels):
        super().__init__(track, output_frequency, output_channels)
        self.__track = track

    def __del__(self):
        super().__del__()
        del self.__track


class SoraVideoSink(SoraVideoSinkImpl):
    def __init__(self, track):
        super().__init__(track)
        self.__track = track

    def __del__(self):
        super().__del__()
        del self.__track
